/* -*- mode: c++ -*-
   this file is part of the robolog libraries
   control interface for the server connection
   Copyright (C) Oct 22 1997 Oliver Obst
   $Id: control.C,v 1.5 1999/05/05 13:46:48 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>     // sprintf

#if STDC_HEADERS
# include <stdlib.h>    // atoi
#endif

#include <sys/types.h> // FD.. macros
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include "control.h"
#include "knowledge.h"
#include "oclient.h"
#include "soccerconst.h"
#include "debug.h"

Control::Control(Listproc *l, const char *server, int port, double ver)
        : OClient(server, port)
{
        sensor = l;
        sensor->set_moved(false);
        sensor->set_version(ver);
        have_data = false;
}

Control::~Control()
{
        close();
}

// --- public -----------------------------------------------------------------
// Control :: init
// send init message
void Control::init(bool goalie)
{
        open();
        if (sensor->get_version() < 4.0)
                sprintf(tmpstr, "(init %s)", sensor->get_teamname());
        else {
                if (goalie) 
                        sprintf(tmpstr,
                                "(init %s (version %f) (goalie))",
                                sensor->get_teamname(), sensor->get_version());
                else
                        sprintf(tmpstr, 
                                "(init %s (version %f))",
                                sensor->get_teamname(), sensor->get_version());
        }
        outq.write(tmpstr);
}        

// --- public -----------------------------------------------------------------
// Control :: moveto
//              double x, y        absolute player position
// for moving the player to position x,y. Origin is the center mark. Positive
// X values are on the opponents side, Y goes towards the right side.
// works only in 'before_kick_off' mode!
void Control::moveto(double x, double y) 
{
        sprintf(tmpstr, "(move %f %f)", x, y);
        outq.write(tmpstr, true);
}

// --- public -----------------------------------------------------------------
// Control :: turn
//              double moment
void Control::turn(double moment) 
{
        if (moment < min_turn_moment)
                moment += 360.0;
        if (moment > max_turn_moment)
                moment -= 360.0;

        if (moment >=  min_turn_moment && moment <= max_turn_moment) {
                sprintf(tmpstr, "(turn %f)", moment);
                outq.write(tmpstr, true);
        }
}

// --- public -----------------------------------------------------------------
// turn the players head
void Control::turn_neck(double angle_d) 
{
        if (angle_d >=  min_turn_moment && angle_d <= max_turn_moment) {
                sprintf(tmpstr, "(turn_neck %f)", angle_d);
                outq.write(tmpstr, false); // this is always allowed
        }
}

// --- public -----------------------------------------------------------------
// Control :: dash
//              double power
void Control::dash(double power) 
{
        if (power < min_dash_power)
                power = min_dash_power;
        if (power > max_dash_power)
                power = max_dash_power;
        
        sprintf(tmpstr, "(dash %f)", power);
        outq.write(tmpstr, true);
}

// --- public -----------------------------------------------------------------
void Control::kick(double power, double direction) 
{
        if (direction < min_kick_moment)
                direction += 360.0;
        if (direction > max_kick_moment)
                direction -= 360.0;
        if (power < min_kick_power)
                power = min_kick_power;
        if (power > max_kick_power)
                power = max_kick_power;

        if (direction >= min_kick_moment && direction <= max_kick_moment) {
                sprintf(tmpstr, "(kick %f %f)", power, direction);
                outq.write(tmpstr, true);
        }
}

// --- public -----------------------------------------------------------------
void Control::say(const char *text)
{
        if (text != 0 && strlen(text) > 0) {
                sprintf(tmpstr, "(say %s)", text);
                outq.write(tmpstr, true);
        }
}

// --- public -----------------------------------------------------------------
// Control :: goalie_catch(double direction)
//              double direction
// can catch ball in catchable_area around the goal
void Control::goalie_catch(double direction)
{
        if (direction < min_catch_moment)
                direction += 360.0;
        if (direction > max_catch_moment)
                direction -= 360.0;
        
        if (direction >= min_catch_moment &&
            direction <= max_catch_moment) {
                sprintf(tmpstr, "(catch %f)", direction);
                outq.write(tmpstr, true);
        }
}

// --- public -----------------------------------------------------------------
// Control :: changeview
//              Angle   width
//              Quality quality
void Control::changeview(Angle width, Quality quality)
{
        char s1[10], s2[10];
        
        switch (width) {
        case normal :
                strcpy(s1, "normal");
                break;
        case wide :
                strcpy(s1, "wide");
                break;
        case narrow :
                strcpy(s1, "narrow");
        }

        if (quality == high)
                strcpy(s2, "high");
        else
                strcpy(s2, "low");
        
        sprintf(tmpstr, "(change_view %s %s)", s1, s2);
        outq.write(tmpstr);        
}

// --- public -----------------------------------------------------------------
// Control :: sense_body()
// request internal player status information
void Control :: sense_body()
{
        if (sensor->get_version() >= 4.0) {
                strcpy(tmpstr, "(sense_body)");
                outq.write(tmpstr);
        }
}

// --- public -----------------------------------------------------------------
// Control :: wait4outq
//              int level
// wait until outq contains less or equal than level entries
void Control :: wait4outq(int level)
{
        while (outq.size() > level);
}

// --- public -----------------------------------------------------------------
// send and receive data to and from server
void Control::transceive()
{
         int     m;
         fd_set  readfds, writefds;
         timeval tv;
         
         FD_ZERO(&readfds);
         FD_SET(sockfd1, &readfds);
         FD_ZERO(&writefds);

         // wait at most 10 ms for incoming data
         tv.tv_sec = 0;
         tv.tv_usec = 10000;
         
         if (!outq.empty() && !(sensor->get_moved() && outq.peek_move()))
                 FD_SET(sockfd1, &writefds);

         select(sockfd1 + 1, &readfds, &writefds, 0, &tv);

         if (FD_ISSET(sockfd1, &readfds)) {
                 if ((m = OClient::receive(netstr, MaxStringSize)) > 0) {
                         memcpy(inq, netstr, m);
                         m < MaxStringSize ?
                                 inq[m] = '\000' : inq[MaxStringSize] = '\000';
                         have_data = true;
                 }
         }
        
         if (FD_ISSET(sockfd1, &writefds)) {
                 bool move;
                 if ((m = outq.read(netstr, &move, false)) > 0) {
                         ROBOBUG1(ROBOBUG_INFO,CONTROL_SEND,
                                  "sending (%s)",netstr);
                         if (move) sensor->set_moved();
                         send(netstr, m);
                 }
         }
}

// --- public -----------------------------------------------------------------
// send data to server
void Control::dispatch()
{
         int     m;
         fd_set  writefds;
         timeval tv;
         
         FD_ZERO(&writefds);

         // don't wait 
         tv.tv_sec = 0;
         tv.tv_usec = 0;
         
         if (!outq.empty() && !(sensor->get_moved() && outq.peek_move()))
                 FD_SET(sockfd1, &writefds);

         select(sockfd1 + 1, 0, &writefds, 0, &tv);

         if (FD_ISSET(sockfd1, &writefds)) {
                 bool move;
                 if ((m = outq.read(netstr, &move, false)) > 0) {
                         ROBOBUG1(ROBOBUG_INFO,CONTROL_SEND,
                                  "sending (%s)",netstr);
                         if (move) sensor->set_moved();
                         send(netstr, m);
                 }
         }
}

// --- public -----------------------------------------------------------------
// receive data from server
void Control::receive()
{
         int     m;
         fd_set  readfds;
         timeval tv;
         
         FD_ZERO(&readfds);
         FD_SET(sockfd1, &readfds);

         // don't wait for incoming data
         tv.tv_sec = 0;
         tv.tv_usec = 0;
         
         select(sockfd1 + 1, &readfds, 0, 0, &tv);

         if (FD_ISSET(sockfd1, &readfds)) {
                 if ((m = OClient::receive(netstr, MaxStringSize)) > 0) {
                         memcpy(inq, netstr, m);
                         m < MaxStringSize ?
                                 inq[m] = '\000' : inq[MaxStringSize] = '\000';
                         have_data = true;
                 }
         }        
}

// --- public -----------------------------------------------------------------
// get data from inq
bool Control::get_data(char *data)
{
        bool done = have_data;

        if (have_data) {
                strcpy(data, inq);
                have_data = false;
        }

        return done;
}

// --- public -----------------------------------------------------------------
// dequeue all commands from the outq
void Control::dequeue()
{
        outq.clear();
}

// --- public -----------------------------------------------------------------
// return the number of commands in the outq
int Control::outqlen()
{
        return outq.size();
}
