/* -*- mode: c -*-
   
   this file is part of the robolog libraries
   Copyright (C) Jul 29 1998 Oliver Obst
   eclipse prolog interface for the robolog c library
   $Id: ec_robolog.c,v 1.6 1999/05/05 13:53:38 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <errno.h>

#include <robolog.h>
#include "ceclipse.h"

/**
 * atan2 for prolog.
 * This function calculates the arc tangent of the two
 * numbers x and y. It is similar to calculating the arc
 * tangent of y / x, except that the signs of both arguments
 * are used to determine the quadrant of the result. (from man atan2(3))
 *
 * Prolog: atan2(+X, +Y, ?Result)
 */
int p_atan2()
{
        double  x;
        double  y;
        int     err = PSUCCEED;
        
        pword py = ec_arg(1);
        pword px = ec_arg(2);
        pword pres;
        
        if (ec_ok(ec_get_number(py, &y), &err) &&
            ec_ok(ec_get_number(px, &x), &err)) {
                pres = ec_double(RSC_atan2(y, x));
                err = ec_unify_arg(3, pres);
        }

        return err;
}
        
/**
 * deg2rad for prolog.
 * Converts a number from degrees to radiants.
 *
 * Prolog: deg2rad(+Deg, ?Rad)
 */
int p_deg2rad()
{
        double  deg;
        int     err = PSUCCEED;
        
        pword pdeg = ec_arg(1);
        pword pres;
        
        if (ec_ok(ec_get_number(pdeg, &deg), &err)) {
                pres = ec_double(RSC_deg2rad(deg));
                err = ec_unify_arg(2, pres);
        }

        return err;
}     

/**
 * rad2deg for prolog.
 * Converts a number from radiants to degrees.
 *
 * Prolog: rad2deg(+Rad, ?Deg)
 */
int p_rad2deg()
{
        double  rad;
        int     err = PSUCCEED;
        
        pword prad = ec_arg(1);
        pword pres;
        
        if (ec_ok(ec_get_number(prad, &rad), &err)) {
                pres = ec_double(RSC_rad2deg(rad));
                err = ec_unify_arg(2, pres);
        }

        return err;
}

/**
 * server registration for normal players.
 *
 * Prolog: init_player(+Teamname, +Servername, +Portnr, +Version)
 */
int p_init_player()
{
        char    *teamname;
        char    *servername;
        long    portnr;
        double  version;
        int     err = PSUCCEED;
        
        pword team   = ec_arg(1);
        pword server = ec_arg(2);
        pword port   = ec_arg(3);
        pword vers   = ec_arg(4);

        if (ec_ok(ec_get_string(team, &teamname), &err) &&
            ec_ok(ec_get_string(server, &servername), &err) &&
            ec_ok(ec_get_long(port, &portnr), &err)  &&
            ec_ok(ec_get_number(vers, &version), &err)) {
                RSC_init_player(teamname, servername, portnr, version);
        }
        return err;
}

/**
 * server registration for goalies.
 *
 * Prolog: init_goalie(+Teamname, +Servername, +Portnr, +Version)
 */
int p_init_goalie()
{
        char    *teamname;
        char    *servername;
        long    portnr;
        double  version;
        int     err = PSUCCEED;
        
        pword team   = ec_arg(1);
        pword server = ec_arg(2);
        pword port   = ec_arg(3);
        pword vers   = ec_arg(4);

        if (ec_ok(ec_get_string(team, &teamname), &err) &&
            ec_ok(ec_get_string(server, &servername), &err) &&
            ec_ok(ec_get_long(port, &portnr), &err)  &&
            ec_ok(ec_get_number(vers, &version), &err)) {
                RSC_init_goalie(teamname, servername, portnr, version);
        }
        return err;
}

/**
 * stops IRQ and deletes client
 *
 * Prolog: stop/0
 */
int p_stop()
{
        RSC_stop();
        return PSUCCEED;
}

/**
 * open_log(+Filename, +Append, +Flushfreq)   
 * start a normal logfile 'Filename', which will be flushed every
 * 'Flushfreq' lines.
 * If 'Append' is false, an existing file 'Filename' will be overwritten.
 */
int p_open_log()
{
        char    *filename;
        dident  app;
        long    freq;
        int     err = PSUCCEED;
        
        pword file   = ec_arg(1);
        pword a      = ec_arg(2);
        pword f      = ec_arg(3);
        
        if (ec_ok(ec_get_string(file, &filename), &err) &&
            ec_ok(ec_get_atom(a, &app), &err) &&
            ec_ok(ec_get_long(f, &freq), &err)) {
                /* FIXME: add append and flush freq. */
                if (RSC_open_log(filename) != 0) {
                        err = SYS_ERROR;
                }
        }
        return err;
}


/**
 * open_audiolog(+Filename, +Append, +Flushfreq)   
 * start an audio logfile 'Filename', which will be flushed every
 * 'Flushfreq' lines.
 * If 'Append' is false, an existing file 'Filename' will be overwritten.
 */
int p_open_audiolog()
{
        char    *filename;
        dident  app;
        long    freq;
        int     err = PSUCCEED;
        
        pword file   = ec_arg(1);
        pword a      = ec_arg(2);
        pword f      = ec_arg(3);

        if (ec_ok(ec_get_string(file, &filename), &err) &&
            ec_ok(ec_get_atom(a, &app), &err) &&
            ec_ok(ec_get_long(f, &freq), &err)) {
                /* FIXME: use all parameters */
                if (RSC_open_audiolog(filename) != 0) {
                        err = SYS_ERROR;
                }
        }
        return err;
}

/**
 * close_log
 */
int p_close_log()
{
        RSC_close_log();
        return PSUCCEED;
}

/**
 * close_audiolog
 */
int p_close_audiolog()
{
        RSC_close_audiolog();
        return PSUCCEED;
}

/**
 * start_viewserver(+Base_port)
 */
int p_start_viewserver()
{
        
        long base;
        pword p = ec_arg(1);
        int     err = PSUCCEED;
        
        if (ec_ok(ec_get_long(p, &base), &err)) {
                if (RSC_start_viewserver(base) != 0) err = PFAIL;
        }
        
        return err;
}

/**
 * stop_viewserver/0
 */
int p_stop_viewserver()
{
        RSC_stop_viewserver();
        return PSUCCEED;
}

/**
 * returns simulation step of actual working space data
 */
int p_simstep()
{
        return ec_unify_arg(1, ec_long(RSC_simstep()));
}

/**
 *
 * returns internal robolog clock
 */
int p_roboclock()
{
        return ec_unify_arg(1, ec_long(RSC_roboclock()));
}


/**
 * refreshs workspace
 */
int p_refresh()
{
        RSC_refresh();
        return PSUCCEED;
}

/**
 * waits for new data and refresh workspace
 */
int p_await_refresh()
{
        RSC_await_refresh();
        return PSUCCEED;
}

/**
 * waits until outq is empty
 */
int p_wait_for_robo()
{
        RSC_wait_for_robo(0);
        return PSUCCEED;
}

/**
 * waits until outq is almost empty
 */
int p_wait_for_robo3()
{
        RSC_wait_for_robo(2);
        return PSUCCEED;
}

/**
 * dequeue/0
 * delete the outq
 */
int p_dequeue()
{
        RSC_dequeue();
        return PSUCCEED;
}

/**
 * queuelen(?Len)
 * return number of commands in the outq
 */
int p_queuelen()
{
        return ec_unify_arg(1, ec_long(RSC_outqlen()));
}

/**
 * teamname(+Team, ?Name)
 * return team name of own/other team (self/other)
 * fails if name unknown
 */
int p_teamname()
{
        int     res = PSUCCEED;
        dident  pteam;
        char    *name;
        
        if (ec_ok(ec_get_atom(ec_arg(1), &pteam), &res)) {
                char *team = ec_did_name(pteam);
                
                if (strcmp(team, "self") == 0) {
                        name = RSC_teamname(0);
                        if (strlen(name) > 0) {
                                res = ec_unify_arg(2, ec_string(name));
                        }
                        else res = PFAIL;
                } else {
                        if (strcmp(team, "other") == 0) {
                                name = RSC_teamname(1);
                                if (strlen(name) > 0) {
                                        res = ec_unify_arg(2, ec_string(name));
                                }
                                else res = PFAIL;
                        }
                        else res = RANGE_ERROR;
                }
        }

        return res;
}

/**
 * returns clients uniform number
 */
int p_uniformnumber() 
{
        return ec_unify_arg(1, ec_long(RSC_uniformnumber()));
}

/**
 * succeed if client is goalie
 */
int p_is_goalie()
{
        if (RSC_is_goalie()) return PSUCCEED;
        else return PFAIL;
}

/**
 * goalie(+Team, ?Nr)
 * returns goalie uniform number of own/other team (self/other)
 * fail if goalie nr unknown
 */
int p_goalie_nr()
{
        int     res = PSUCCEED;
        int     nr;
        dident  pteam;
        
        if (ec_ok(ec_get_atom(ec_arg(1), &pteam), &res)) {
                char *team = ec_did_name(pteam);
                
                if (strcmp(team, "self") == 0) {
                        nr = RSC_goalie(0);
                        if (nr > 0) res = ec_unify_arg(2, ec_long(nr));
                        else res = PFAIL;
                } else {
                        if (strcmp(team, "other") == 0) {
                                nr = RSC_goalie(1);
                                if (nr > 0) res = ec_unify_arg(2, ec_long(nr));
                                else res = PFAIL;
                        }
                        else res = RANGE_ERROR;
                }
        }

        return res;
}

/**
 * referee(?Status)
 * returns referee judgement
 */
int p_referee()
{
        pword pstatus;
        
        switch (RSC_referee()) {
        case st_before_kick_off :
                pstatus = ec_atom(ec_did("before_kick_off", 0));
                break;
        case st_kick_off_l :
                pstatus = ec_atom(ec_did("kick_off_l", 0));
                break;
        case st_kick_off_r :
                pstatus = ec_atom(ec_did("kick_off_r", 0));
                break;
        case st_kick_in_l :
                pstatus = ec_atom(ec_did("kick_in_l", 0));
                break;
        case st_kick_in_r :
                pstatus = ec_atom(ec_did("kick_in_r", 0));
                break;
        case st_corner_kick_l :
                pstatus = ec_atom(ec_did("corner_kick_l", 0));
                break;
        case st_corner_kick_r :
                pstatus = ec_atom(ec_did("corner_kick_r", 0));
                break;
        case st_goal_kick_l :
                pstatus = ec_atom(ec_did("goal_kick_l", 0));
                break;
        case st_goal_kick_r :
                pstatus = ec_atom(ec_did("goal_kick_r", 0));
                break;
        case st_free_kick_l :
                pstatus = ec_atom(ec_did("free_kick_l", 0));
                break;
        case st_free_kick_r :
                pstatus = ec_atom(ec_did("free_kick_r", 0));
                break;
        case st_play_on :
                pstatus = ec_atom(ec_did("play_on", 0));
                break;
        case st_half_time :
                pstatus = ec_atom(ec_did("half_time", 0));
                break;
        case st_time_up :
                pstatus = ec_atom(ec_did("time_over", 0));
                break;
        case st_extend :
                pstatus = ec_atom(ec_did("extend", 0));
                break;
        case st_foul_l_ :
                pstatus = ec_atom(ec_did("foul_l", 0));
                break;
        case st_foul_r_ :
                pstatus = ec_atom(ec_did("foul_r", 0));
                break;
        case st_goal_l_ :
                pstatus = ec_atom(ec_did("goal_l", 0));
                break;
        case st_goal_r_ :
                pstatus = ec_atom(ec_did("goal_r", 0));
                break;
        case st_goalie_catch_ball_l :
                pstatus = ec_atom(ec_did("goalie_catch_ball_l", 0));
                break;
        case st_goalie_catch_ball_r :
                pstatus = ec_atom(ec_did("goalie_catch_ball_r", 0));
                break;
        default :
                pstatus = ec_atom(ec_did("undefined", 0));
        }

        return ec_unify_arg(1, pstatus);
}

/**
 * myside(?Side)
 * returns side the client plays on
 */
int p_myside()
{
        int res;

        switch (RSC_myside()) {
        case left_side :
                res = ec_unify_arg(1, ec_atom(ec_did("l", 0)));
                break;
        case right_side :
                res = ec_unify_arg(1, ec_atom(ec_did("r", 0)));
                break;
        default :
                res = PFAIL;
        }
        
        return res;        
}

/**
 * 
 * returns opponents side
 */
int p_goal_side()
{
        int res;

        switch (RSC_myside()) {
        case left_side :
                res = ec_unify_arg(1, ec_atom(ec_did("r", 0)));
                break;
        case right_side :
                res = ec_unify_arg(1, ec_atom(ec_did("l", 0)));
                break;
        default :
                res = PFAIL;
        }
        
        return res;
}

/**
 * returns clients position
 *
 * Prolog self(+Time, ?X, ?Y, ?BodyDir, ?Facedir)
 */
int p_self()
{
        pword arg1 = ec_arg(1);
        pword arg2 = ec_arg(2);
        pword arg3 = ec_arg(3);
        pword arg4 = ec_arg(4);
        pword arg5 = ec_arg(5);

        long int time;
        int res = PSUCCEED;
        
        double xpos, ypos, angle1, angle2;
        
        if (ec_ok(ec_get_long(arg1, &time), &res)) {
                RSC_self(time, &xpos, &ypos, &angle1, &angle2, GET);
                arg2 = ec_double(xpos);
                arg3 = ec_double(ypos);
                arg4 = ec_double(angle1);
                arg5 = ec_double(angle2);
                ec_ok(ec_unify_arg(2, arg2), &res);
                ec_ok(ec_unify_arg(3, arg3), &res);
                ec_ok(ec_unify_arg(4, arg4), &res);
                ec_ok(ec_unify_arg(5, arg5), &res);
        }
        
        return res;
}

/*** maths line stuff ********************************************************/

/**
 * point_on_line(+Px,+Py,+X0,+Y0,+X1,+Y1)
 */
int p_point_on_line()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                if (!RSC_point_on_line(px,py,x0,y0,x1,y1))
                        res = PFAIL;
        }
        return res;
}

/**
 * point_between(+Px,+Py,+X0,+Y0,+X1,+Y1)
 */
int p_point_between()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                if (!RSC_point_between(px,py,x0,y0,x1,y1))
                        res = PFAIL;
        }
        return res;
}

/**
 * dist_point_line(+Px,+Py,+X0,+Y0,+X1,+Y1,?Dist)
 */
int p_dist_point_line()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1, dist;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                dist = RSC_dist_point_line(px,py,x0,y0,x1,y1);
                res = ec_unify_arg(7, ec_double(dist));
        }
        return res;
}

/**
 * line_angle(+X0,+Y0,+X1,+Y1,?Angle)
 */
int p_line_angle()
{
        int    res = PSUCCEED;
        double x0, y0, x1, y1, angle;
        if (ec_ok(ec_get_number(ec_arg(1), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y1), &res)) {
                angle = RSC_line_angle(x0,y0,x1,y1);
                res = ec_unify_arg(5, ec_double(angle));
        }
        return res;
}

/*** maths rectangle stuff ***************************************************/

/**
 * point_in_box(+Px,+Py,+X0,+Y0,+X1,+Y1)
 */
int p_point_in_box()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                if (!RSC_point_in_box(px,py,x0,y0,x1,y1))
                        res = PFAIL;
        }
        return res;
}

/**
 * nearest_edge(+Px,+Py,+X0,+Y0,+X1,+Y1,?X2,?Y2)
 */
int p_nearest_edge()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1, x2, y2;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                RSC_nearest_edge(px,py,x0,y0,x1,y1,&x2,&y2);
                ec_ok(ec_unify_arg(7, ec_double(x2)), &res);
                ec_ok(ec_unify_arg(8, ec_double(y2)), &res);
        }
        return res;
}

/**
 * distance_to_edge(+Px,+Py,+X0,+Y0,+X1,+Y1,?Dist)
 */
int p_distance_to_edge()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1, x2, y2;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                double dist = RSC_distance_to_edge(px,py,x0,y0,x1,y1);
                ec_unify_arg(7, ec_double(dist));
        }
        return res;
}

/**
 * closest_point_inside(+Px,+Py,+X0,+Y0,+X1,+Y1,?X2,?Y2)
 */
int p_closest_point_inside()
{
        int    res = PSUCCEED;
        double px, py, x0, y0, x1, y1, x2, y2;
        if (ec_ok(ec_get_number(ec_arg(1), &px), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &py), &res) &&
            ec_ok(ec_get_number(ec_arg(3), &x0), &res) &&
            ec_ok(ec_get_number(ec_arg(4), &y0), &res) &&
            ec_ok(ec_get_number(ec_arg(5), &x1), &res) &&
            ec_ok(ec_get_number(ec_arg(6), &y1), &res)) {
                RSC_closest_point_inside(px,py,x0,y0,x1,y1,&x2,&y2);
                ec_ok(ec_unify_arg(7, ec_double(x2)), &res);
                ec_ok(ec_unify_arg(8, ec_double(y2)), &res);
        }
        return res;
}

/**
 * translate direction number into an atom
 */
dident h_dir_quality(int ndir)
{
        dident ddir;

        switch (ndir) {
        case 0 :
                ddir = ec_did("n", 0);
                break;
        case 1 :
                ddir = ec_did("ne", 0);
                break;
        case 2 :
                ddir = ec_did("e", 0);
                break;
        case 3 :
                ddir = ec_did("se", 0);
                break;
        case 4 :
                ddir = ec_did("s", 0);
                break;
        case 5 :
                ddir = ec_did("sw", 0);
                break;
        case 6 :
                ddir = ec_did("w", 0);
                break;
        case 7 :
                ddir = ec_did("nw", 0);
                break;
        default:
                ddir = ec_did("unknown", 0);
        }
        return ddir;
}

/**
 * translate distance number into atom
 */
dident h_dist_quality(int ndist)
{
        dident dist;
        
        switch (ndist) {
        case 0 :
                dist = ec_did("close", 0);
                break;
        case 1 :
                dist = ec_did("near", 0);
                break;
        case 2 :
                dist = ec_did("short", 0);
                break;
        case 3 :
                dist = ec_did("far", 0);
                break;
        case 4 :
                dist = ec_did("remote", 0);
                break;
        default :
                dist = ec_did("unknown", 0);
        }

        return dist;
}

/**
 * translate quality atom into number
 */
int h_quality_n(dident dq, int *n)
{
        int m;
        char *q = ec_did_name(dq);
        if ((m = (strlen(q) > 2))) {
                if (!strcmp(q, "close")) *n = 0;
                else if (!strcmp(q, "near")) *n = 1;
                else if (!strcmp(q, "short")) *n = 2;
                else if (!strcmp(q, "far")) *n = 3;
                else if (!strcmp(q, "remote")) *n = 4;
                else *n = -1;
        } else {
                if (!strcmp(q, "n")) *n = 0;
                else if (!strcmp(q, "ne")) *n = 1;
                else if (!strcmp(q, "e")) *n = 2;
                else if (!strcmp(q, "se")) *n = 3;
                else if (!strcmp(q, "s")) *n = 4;
                else if (!strcmp(q, "sw")) *n = 5;
                else if (!strcmp(q, "w")) *n = 6;
                else if (!strcmp(q, "nw")) *n = 7;
                else *n = -1;
        }
        return m;
}

/**
 * translate angle into a compass direction
 * (n,s,e,w,ne,nw,se,sw)
 * Prolog: dir_quality(+Angle, ?Direction)
 */
int p_dir_quality()
{
        int    res = PSUCCEED;
        dident dir;
        
        double angle;
        
        if (ec_ok(ec_get_number(ec_arg(1), &angle), &res)) {
                dir = h_dir_quality(RSC_dir_quality(angle));
                res = ec_unify_arg(2, ec_atom(dir));
        }
        return res;
}

/**
 * translate distance quantity into distance quality
 * 
 * Prolog: dist_quality(+Dist, ?Quality)
 */
int p_dist_quality()
{
        int    res = PSUCCEED;
        dident dqual;
        
        double dist;
        
        if (ec_ok(ec_get_number(ec_arg(1), &dist), &res)) {
                dqual = h_dist_quality(RSC_dist_quality(dist));
                res = ec_unify_arg(2, ec_atom(dqual));
        }

        return res;
}

/**
 * pred. quality
 *
 * Prolog: pred_quality(+Quality0, ?Quality1)
 */
int p_pred_quality()
{
        int     res = PSUCCEED;
        dident  pqual;
        int n0, n1;

        if (ec_ok(ec_get_atom(ec_arg(1), &pqual), &res)) {
                if (h_quality_n(pqual, &n0)) {
                        n1 = RSC_pred_quality(n0, 4, 0);
                        pqual = h_dist_quality(n1);
                } else {
                        n1 = RSC_pred_quality(n0, 7, 1);
                        pqual = h_dir_quality(n1);
                }
                res = ec_unify_arg(2, ec_atom(pqual));
        }
        return res;
}

/**
 * succ. quality
 *
 * Prolog: succ_quality(+Quality0, ?Quality1)
 */
int p_succ_quality()
{
        int     res = PSUCCEED;
        dident  pqual;
        int n0, n1;

        if (ec_ok(ec_get_atom(ec_arg(1), &pqual), &res)) {
                if (h_quality_n(pqual, &n0)) {
                        n1 = RSC_succ_quality(n0, 4, 0);
                        pqual = h_dist_quality(n1);
                } else {
                        n1 = RSC_succ_quality(n0, 7, 1);
                        pqual = h_dir_quality(n1);
                }
                res = ec_unify_arg(2, ec_atom(pqual));
        }
        return res;
}

/**
 * opposite quality
 *
 * Prolog: opposite(+DirQuality0, ?DirQuality1)
 */
int p_opposite()
{
        int     res = PSUCCEED;
        dident  pqual;
        int n0, n1;

        if (ec_ok(ec_get_atom(ec_arg(1), &pqual), &res)) {
                if (h_quality_n(pqual, &n0)) {
                        pqual = h_dir_quality(-1);
                } else {
                        n1 = (n0 < 4 ? n0 + 4 : n0 - 4);
                        pqual = h_dir_quality(n1);
                }
                
                res = ec_unify_arg(2, ec_atom(pqual));
        }
        return res;
}
        

/**
 * calculates distance to (X,Y), fails if own position unknown
 *
 * Prolog: dist(+X, +Y, -Dist)
 */
int p_dist()
{
        pword arg1 = ec_arg(1);
        pword arg2 = ec_arg(2);
        pword arg3;

        int    res = PSUCCEED;

        double x1, y1, dist;
        
        if (ec_ok(ec_get_number(arg1, &x1), &res)
            && ec_ok(ec_get_number(arg2, &y1), &res)) {
                if (RSC_dist(x1, y1, &dist)) {
                        arg3 = ec_double(dist);
                        ec_ok(ec_unify_arg(3, arg3), &res);
                }
                else res = PFAIL;
        }

        return res;
}

/**
 * gets absolute angle between heading and the line between robot and (X,Y)
 *
 * Prolog: absangle(+X,+Y,?A)
 */
int p_absangle()
{
        pword arg1 = ec_arg(1);
        pword arg2 = ec_arg(2);
        pword arg3;

        int    res = PSUCCEED;

        double x1, y1, angle;
        
        if (ec_ok(ec_get_number(arg1, &x1), &res)
            && ec_ok(ec_get_number(arg2, &y1), &res)) {
                if (RSC_absangle(x1, y1, &angle)) {
                        arg3 = ec_double(angle);
                        ec_ok(ec_unify_arg(3, arg3), &res);
                }
                else res = PFAIL;
        }

        return res;
}

/**
 * sets/gets view width
 *
 * Prolog: view(?Width)
 */
int p_view_1()
{
        int     res = PSUCCEED;
        pword   arg = ec_arg(1);
        dident  pwidth;
        
        int angle;

        if (ec_ok(ec_get_atom(arg, &pwidth), &res)) {
                char *width = ec_did_name(pwidth);
                
                if (strcmp(width, "normal") == 0) {
                        angle = v_normal;
                };
                
                if (strcmp(width, "wide") == 0) {
                        angle = v_wide;
                };
                        
                if (strcmp(width, "narrow") == 0) {
                        angle = v_narrow;
                };

                RSC_view_1(&angle, SET);
        } else {
                RSC_view_1(&angle, GET);
                switch (angle) {
                case v_normal :
                        pwidth = ec_did("normal", 0);
                        break;
                case v_wide :
                        pwidth = ec_did("wide", 0);
                        break;
                case v_narrow :
                        pwidth = ec_did("narrow", 0);
                }
                
                res = ec_unify_arg(1, ec_atom(pwidth));
        }

        return res;
}

/**
 * sets/gets view width and quality
 */
int p_view_2()
{
        
        int     res1 = PSUCCEED;
        int     res2 = PSUCCEED;
        pword   arg1 = ec_arg(1);
        pword   arg2 = ec_arg(2);
        
        dident  pangle;
        dident  pquality;
        
        int angle;
        int quality;

        if (ec_ok(ec_get_atom(arg1, &pangle), &res1)) {
                char *sangle = ec_did_name(pangle);
                
                if (strcmp(sangle, "normal") == 0) {
                        angle = v_normal;
                };
                
                if (strcmp(sangle, "wide") == 0) {
                        angle = v_wide;
                };      
                        
                if (strcmp(sangle, "narrow") == 0) {
                        angle = v_narrow;
                };

                RSC_view_1(&angle, SET);
        } else {
                RSC_view_1(&angle, GET);
                switch (angle) {
                case v_normal :
                        pangle = ec_did("normal", 0);
                        break;
                case v_wide :
                        pangle = ec_did("wide", 0);
                        break;
                case v_narrow :
                        pangle = ec_did("narrow", 0);
                }                
                res1 = ec_unify_arg(1, ec_atom(pangle));
        }

        if (ec_ok(ec_get_atom(arg2, &pquality), &res2)) {
                char *squality = ec_did_name(pquality);
                
                if (strcmp(squality, "high") == 0) {
                        quality = v_high;
                };
                
                if (strcmp(squality, "low") == 0) {
                        quality = v_low;
                };

                RSC_view_2(&angle, &quality, GET, SET);
        } else {
                RSC_view_2(&angle, &quality, GET, GET);
                switch (quality) {
                case v_high :
                        pquality = ec_did("high", 0);
                        break;
                case v_low :
                        pquality = ec_did("low", 0);
                }                
                res2 = ec_unify_arg(2, ec_atom(pquality));
        }        
        ec_ok(res2, &res1);

        return res1;
}

/**
 * get current messages index
 *
 * Prolog: msgidx(-Index)
 */
int p_msgidx()
{
        return ec_unify_arg(1, ec_long(RSC_msgidx()));
}

/**
 * get next messages index
 *
 * Prolog: nextmsgidx(-Index)
 */
int p_nextmsgidx()
{
        int res;
        long int idx = RSC_nextmsgidx();

        if (idx < 0)
                res = PFAIL;
        else
                res = ec_unify_arg(1, ec_long(idx));

        return res;
}

/**
 * get number of messages of sender s at simulation step t
 *      s is self, referee, other or all
 *
 * Prolog: messages(+Time, +Sender, ?Number)
 */
int p_n_message()
{
        long    n, t;
        int     res = ec_get_long(ec_arg(1), &t);
        pword   arg2 = ec_arg(2);
        dident  psender;
        
        if (ec_ok(ec_get_atom(arg2, &psender), &res)) {
                char *sender = ec_did_name(psender);

                if (strcmp(sender, "other") == 0) {
                        n = RSC_n_message(2, t);
                        res = ec_unify_arg(3, ec_long(n));
                }
                else if (strcmp(sender, "referee") == 0) {
                        n = RSC_n_message(1, t);
                        res = ec_unify_arg(3, ec_long(n));
                }
                else if (strcmp(sender, "self") == 0) {
                        n = RSC_n_message(0, t);
                        res = ec_unify_arg(3, ec_long(n));
                }
                else if (strcmp(sender, "all") == 0) {
                        n = RSC_n_message(3, t);
                        res = ec_unify_arg(3, ec_long(n));
                }
                else res = RANGE_ERROR;
        }
        
        return res;
}

/**
 * get message i of sender s at simulation step t
*      s is self, referee, other, all
 *
 * Prolog: message(+Time, +Sender, +Index, ?Message, ?Direction)
 */
int p_message()
{
        long    i, t;
        pword   arg2 = ec_arg(2);
        pword   arg3 = ec_arg(3);
        dident  psender;

        char    msg[MaxStringSize];
        double  dir;

        int     res = ec_get_long(arg3, &i);
        ec_ok(ec_get_long(ec_arg(1), &t), &res);

        if (ec_ok(ec_get_atom(arg2, &psender), &res)) {
                char *sender = ec_did_name(psender);
                
                if (strcmp(sender, "other") == 0) {
                        RSC_message(2,t,i,(char *) msg, &dir);
                        ec_ok(ec_unify_arg(4, ec_string(msg)), &res);
                        ec_ok(ec_unify_arg(5, ec_double(dir)), &res);
                }
                else if (strcmp(sender, "referee") == 0) {
                        RSC_message(1,t,i,(char *) msg, &dir);
                        ec_ok(ec_unify_arg(4, ec_string(msg)), &res);
                        ec_ok(ec_unify_arg(5, ec_double(dir)), &res);
                }
                else if (strcmp(sender, "self") == 0) {
                        RSC_message(0,t,i,(char *) msg, &dir);
                        ec_ok(ec_unify_arg(4, ec_string(msg)), &res);
                        ec_ok(ec_unify_arg(5, ec_double(dir)), &res);
                }
                else if (strcmp(sender, "all") == 0) {
                        RSC_message(3,t,i,(char *) msg, &dir);
                        ec_ok(ec_unify_arg(4, ec_string(msg)), &res);
                        ec_ok(ec_unify_arg(5, ec_double(dir)), &res);
                }                
                else res = RANGE_ERROR;
        }
        
        return res;
}

/**
 * returns number of goals seen
 *
 * Prolog: goal(+Time, -Number)
 */
int p_n_goal()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_goal(time))), &res);

        return res;
}

/**
 * gets goal positions
 *
 * Prolog: b_goal(+Time, ?Side,
 *              pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_goal() 
{
        long int i, time;
        int     res = ec_get_long(ec_arg(4), &i);
        object_data_t od;
        dident  pside;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(4, next.val, next.tag);
                i--;

                if (i < 2) {
                        switch (i) {
                        case id_goal_left :
                                RSC_goal(time, id_goal_left, &od);
                                pside = ec_did("l", 0);
                                break;
                        case id_goal_right :
                                RSC_goal(time, id_goal_right, &od);
                                pside = ec_did("r", 0);
                        }
                        res = ec_unify_arg(2, ec_atom(pside));
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns number of balls seen
 *
 * Prolog: ball(+Time, -Number)
 */
int p_n_ball()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_ball(time))), &res);

        return res;
}

/**
 * gets ball position
 *
 * Prolog: ball(+Time, ?Id,
 *              pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_ball()
{
        long int i, time;
        int res = ec_get_long(ec_arg(4), &i);
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                RSC_ball(time, &od);

                ec_ok(ec_unify_arg(2, ec_long(0)), &res);
                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }
        
        return res;
}

/**
 * returns number of flags seen
 *
 * Prolog: flag(+Time, -Number)
 */
int p_n_flag()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_flag(time))), &res);

        return res;
}

/**
 * gets flag position
 *
 * Prolog: flag(+Time, ?Id,
 *              pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_flag() 
{
        long int i, time;
        int     res = ec_get_long(ec_arg(4), &i);
        object_data_t od;
        dident  pflag;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                int found;
                pword next = ec_long(i + 1);
                res = ec_remember(4, next.val, next.tag);
                i--;

                if (i < 53) {
                        switch (i) {
                        case id_flag_l_t :
                                found = RSC_flag(time, id_flag_l_t, &od);
                                pflag = ec_did("flag_l_t", 0);
                                break;
                        case id_flag_l_b :
                                found = RSC_flag(time, id_flag_l_b, &od);
                                pflag = ec_did("flag_l_b", 0);
                                break;
                        case id_flag_c_t :
                                found = RSC_flag(time, id_flag_c_t, &od);
                                pflag = ec_did("flag_c_t", 0);
                                break;
                        case id_flag_c_b :
                                found = RSC_flag(time, id_flag_c_b, &od);
                                pflag = ec_did("flag_c_b", 0);
                                break;
                        case id_flag_r_t :
                                found = RSC_flag(time, id_flag_r_t, &od);
                                pflag = ec_did("flag_r_t", 0);
                                break;
                        case id_flag_r_b :
                                found = RSC_flag(time, id_flag_r_b, &od);
                                pflag = ec_did("flag_r_b", 0);
                                break;
                        case id_pflag_l_t :
                                found = RSC_flag(time, id_pflag_l_t, &od);
                                pflag = ec_did("pflag_l_t", 0);
                                break;
                        case id_pflag_l_c :
                                found = RSC_flag(time, id_pflag_l_c, &od);
                                pflag = ec_did("pflag_l_c", 0);
                                break;
                        case id_pflag_l_b :
                                found = RSC_flag(time, id_pflag_l_b, &od);
                                pflag = ec_did("pflag_l_b", 0);
                                break;
                        case id_pflag_r_t :
                                found = RSC_flag(time, id_pflag_r_t, &od);
                                pflag = ec_did("pflag_r_t", 0);
                                break;
                        case id_pflag_r_c :
                                found = RSC_flag(time, id_pflag_r_c, &od);
                                pflag = ec_did("pflag_r_c", 0);
                                break;
                        case id_pflag_r_b :
                                found = RSC_flag(time, id_pflag_r_b, &od);
                                pflag = ec_did("pflag_r_b", 0);
                                break;
                        case id_flag_c :
                                found = RSC_flag(time, id_flag_c, &od);
                                pflag = ec_did("flag_c", 0);
                                break;
                        case id_gflag_l_t :
                                found = RSC_flag(time, id_gflag_l_t, &od);
                                pflag = ec_did("gflag_l_t", 0);
                                break;
                        case id_gflag_l_b :
                                found = RSC_flag(time, id_gflag_l_b, &od);
                                pflag = ec_did("gflag_l_b", 0);
                                break;
                        case id_gflag_r_t :
                                found = RSC_flag(time, id_gflag_r_t, &od);
                                pflag = ec_did("gflag_r_t", 0);
                                break;
                        case id_gflag_r_b :
                                found = RSC_flag(time, id_gflag_r_b, &od);
                                pflag = ec_did("gflag_r_b", 0);
                                break;
                        case id_oflag_t_l_10 :
                                found = RSC_flag(time, id_oflag_t_l_10, &od);
                                pflag = ec_did("oflag_t_l_10", 0);
                                break;
                        case id_oflag_t_l_20 :
                                found = RSC_flag(time, id_oflag_t_l_20, &od);
                                pflag = ec_did("oflag_t_l_20", 0);
                                break;
                        case id_oflag_t_l_30 :
                                found = RSC_flag(time, id_oflag_t_l_30, &od);
                                pflag = ec_did("oflag_t_l_30", 0);
                                break;
                        case id_oflag_t_l_40 :
                                found = RSC_flag(time, id_oflag_t_l_40, &od);
                                pflag = ec_did("oflag_t_l_40", 0);
                                break;
                        case id_oflag_t_l_50 :
                                found = RSC_flag(time, id_oflag_t_l_50, &od);
                                pflag = ec_did("oflag_t_l_50", 0);
                                break;
                        case id_oflag_t_r_10 :
                                found = RSC_flag(time, id_oflag_t_r_10, &od);
                                pflag = ec_did("oflag_t_r_10", 0);
                                break;
                        case id_oflag_t_r_20 :
                                found = RSC_flag(time, id_oflag_t_r_20, &od);
                                pflag = ec_did("oflag_t_r_20", 0);
                                break;
                        case id_oflag_t_r_30 :
                                found = RSC_flag(time, id_oflag_t_r_30, &od);
                                pflag = ec_did("oflag_t_r_30", 0);
                                break;
                        case id_oflag_t_r_40 :
                                found = RSC_flag(time, id_oflag_t_r_40, &od);
                                pflag = ec_did("oflag_t_r_40", 0);
                                break;
                        case id_oflag_t_r_50 :
                                found = RSC_flag(time, id_oflag_t_r_50, &od);
                                pflag = ec_did("oflag_t_r_50", 0);
                                break;
                        case id_oflag_b_l_10 :
                                found = RSC_flag(time, id_oflag_b_l_10, &od);
                                pflag = ec_did("oflag_b_l_10", 0);
                                break;
                        case id_oflag_b_l_20 :
                                found = RSC_flag(time, id_oflag_b_l_20, &od);
                                pflag = ec_did("oflag_b_l_20", 0);
                                break;
                        case id_oflag_b_l_30 :
                                found = RSC_flag(time, id_oflag_b_l_30, &od);
                                pflag = ec_did("oflag_b_l_30", 0);
                                break;
                        case id_oflag_b_l_40 :
                                found = RSC_flag(time, id_oflag_b_l_40, &od);
                                pflag = ec_did("oflag_b_l_40", 0);
                                break;
                        case id_oflag_b_l_50 :
                                found = RSC_flag(time, id_oflag_b_l_50, &od);
                                pflag = ec_did("oflag_b_l_50", 0);
                                break;
                        case id_oflag_b_r_10 :
                                found = RSC_flag(time, id_oflag_b_r_10, &od);
                                pflag = ec_did("oflag_b_r_10", 0);
                                break;
                        case id_oflag_b_r_20 :
                                found = RSC_flag(time, id_oflag_b_r_20, &od);
                                pflag = ec_did("oflag_b_r_20", 0);
                                break;
                        case id_oflag_b_r_30 :
                                found = RSC_flag(time, id_oflag_b_r_30, &od);
                                pflag = ec_did("oflag_b_r_30", 0);
                                break;
                        case id_oflag_b_r_40 :
                                found = RSC_flag(time, id_oflag_b_r_40, &od);
                                pflag = ec_did("oflag_b_r_40", 0);
                                break;
                        case id_oflag_b_r_50 :
                                found = RSC_flag(time, id_oflag_b_r_50, &od);
                                pflag = ec_did("oflag_b_r_50", 0);
                                break;
                        case id_oflag_l_t_10 :
                                found = RSC_flag(time, id_oflag_l_t_10, &od);
                                pflag = ec_did("oflag_l_t_10", 0);
                                break;
                        case id_oflag_l_t_20 :
                                found = RSC_flag(time, id_oflag_l_t_20, &od);
                                pflag = ec_did("oflag_l_t_20", 0);
                                break;
                        case id_oflag_l_t_30 :
                                found = RSC_flag(time, id_oflag_l_t_30, &od);
                                pflag = ec_did("oflag_l_t_30", 0);
                                break;
                        case id_oflag_l_b_10 :
                                found = RSC_flag(time, id_oflag_l_b_10, &od);
                                pflag = ec_did("oflag_l_b_10", 0);
                                break;
                        case id_oflag_l_b_20 :
                                found = RSC_flag(time, id_oflag_l_b_20, &od);
                                pflag = ec_did("oflag_l_b_20", 0);
                                break;
                        case id_oflag_l_b_30 :
                                found = RSC_flag(time, id_oflag_l_b_30, &od);
                                pflag = ec_did("oflag_l_b_30", 0);
                                break;
                        case id_oflag_r_t_10 :
                                found = RSC_flag(time, id_oflag_r_t_10, &od);
                                pflag = ec_did("oflag_r_t_10", 0);
                                break;
                        case id_oflag_r_t_20 :
                                found = RSC_flag(time, id_oflag_r_t_20, &od);
                                pflag = ec_did("oflag_r_t_20", 0);
                                break;
                        case id_oflag_r_t_30 :
                                found = RSC_flag(time, id_oflag_r_t_30, &od);
                                pflag = ec_did("oflag_r_t_30", 0);
                                break;
                        case id_oflag_r_b_10 :
                                found = RSC_flag(time, id_oflag_r_b_10, &od);
                                pflag = ec_did("oflag_r_b_10", 0);
                                break;
                        case id_oflag_r_b_20 :
                                found = RSC_flag(time, id_oflag_r_b_20, &od);
                                pflag = ec_did("oflag_r_b_20", 0);
                                break;
                        case id_oflag_r_b_30 :
                                found = RSC_flag(time, id_oflag_r_b_30, &od);
                                pflag = ec_did("oflag_r_b_30", 0);
                                break;
                        case id_oflag_t_0 :
                                found = RSC_flag(time, id_oflag_t_0, &od);
                                pflag = ec_did("oflag_t_0", 0);
                                break;
                        case id_oflag_b_0 :
                                found = RSC_flag(time, id_oflag_b_0, &od);
                                pflag = ec_did("oflag_b_0", 0);
                                break;
                        case id_oflag_l_0 :
                                found = RSC_flag(time, id_oflag_l_0, &od);
                                pflag = ec_did("oflag_l_0", 0);
                                break;
                        case id_oflag_r_0 :
                                found = RSC_flag(time, id_oflag_r_0, &od);
                                pflag = ec_did("oflag_r_0", 0);
                        }

                        res = ec_unify_arg(2, ec_atom(pflag));
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns number of flags seen
 *
 * Prolog: flag(+Time, -Number)
 */
int p_n_line()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_line(time))), &res);

        return res;
}

/**
 * gets line position
 *
 * Prolog: line(+Time, ?Id,
 *              pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_line() 
{
        long int i, time;
        int     res = ec_get_long(ec_arg(4), &i);
        object_data_t od;
        dident  pline;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                int found;
                pword next = ec_long(i + 1);
                res = ec_remember(4, next.val, next.tag);
                i--;

                if (i < 5) {
                        switch (i) {
                        case id_line_l :
                                found = RSC_line(time, id_line_l, &od);
                                pline = ec_did("line_l", 0);
                                break;
                        case id_line_r :
                                found = RSC_line(time, id_line_r, &od);
                                pline = ec_did("line_r", 0);
                                break;
                        case id_line_t :
                                found = RSC_line(time, id_line_t, &od);
                                pline = ec_did("line_t", 0);
                                break;
                        case id_line_b :
                                found = RSC_line(time, id_line_b, &od);
                                pline = ec_did("line_b", 0);
                                break;
                        }

                        if (found)
                                res = ec_unify_arg(2, ec_atom(pline));
                        else res = PFAIL;
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns neck angle
 *
 * Prolog: neck(+Time, -Angle)
 */
int p_neck()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_double(RSC_neck(time))), &res);

        return res;
}

/**
 * returns number of team members seen
 *
 * Prolog: partner(+Time, -Number)
 */
int p_n_partner()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_partner(time))), &res);

        return res;
}

/**
 * returns positions of partner with index number 'index'
 * (0 <= index < Number of team members seen)
 * fails if index >= Number of team members seen
 *
 * Prolog: partner (+Time, ?Id,
 *                  pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_partner()
{
        long int i, time;
        int   res = ec_get_long(ec_arg(4), &i);
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(4, next.val, next.tag);
                i--;
                
                if (i < RSC_n_partner(time)) {
                        RSC_partner(time, i, &od);
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(2, ec_long(od.id)), &res);
                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns number of opponents seen
 *
 * Prolog: opponent(-Number)
 */
int p_n_opponent()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_opponent(time))), &res);

        return res;
}

/**
 * returns positions of opponent with index number 'index'
 * (0 <= index < Number of opponents seen)
 * fails if index >= Number of opponents seen
 *
 * Prolog: b_opponent (+Time, ?Id,
 *                  pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_opponent()
{
        long int i, time;
        int   res = ec_get_long(ec_arg(4), &i);
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(4, next.val, next.tag);
                i--;
                
                if (i < RSC_n_opponent(time)) {
                        RSC_opponent(time, i, &od);
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(2, ec_long(od.id)), &res);
                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns number of unidentified players seen
 *
 * Prolog: player(+Time, -Number)
 */
int p_n_player()
{
        long int time;
        
        int res = ec_get_long(ec_arg(1), &time);
        ec_ok(ec_unify_arg(2, ec_long(RSC_n_player(time))), &res);

        return res;
}

/**
 * returns positions of unidentified player with index number 'index'
 * (0 <= index < Number of unidentified players seen)
 * fails if index >= Number of unidentified players seen
 *
 * Prolog: b_player (+Time, ?Id,
 *                   pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_player()
{
        long int i, time;
        int   res = ec_get_long(ec_arg(4), &i);
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(1), &time), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(4, next.val, next.tag);
                i--;
                
                if (i < RSC_n_player(time)) {
                        RSC_player(time, i, &od);
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(2, ec_long(od.id)), &res);
                ec_ok(ec_unify_arg(3, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * sends 'sense_body' command to soccer server. 
 *
 * Prolog: sense_body
 */
int p_sense_body_0()
{
        RSC_sense_body_0();
        return PSUCCEED;
}

/**
 * waits until outq is empty, sends a 'sense_body' command to the server
 * and waits for the result.
 *
 * Prolog: sense_body(-View_width, -View_quality, -Stamina, -Effort,
 *                    -Speed, -Kicks, -Dashs, -Turns, -Says)
 */
int p_sense_body_9()
{
        int res = PSUCCEED;
        int vwidth, vquality;
        long kicks, dashs, turns, says;
        double speed;
        double stamina;
        double effort;
        
        dident pangle;
        dident pquality;

        RSC_sense_body_9(&vwidth, &vquality, &stamina, &effort,
                       &speed, &kicks, &dashs, &turns, &says);

        switch (vwidth) {
        case v_normal :
                pangle = ec_did("normal", 0);
                break;
        case v_wide :
                pangle = ec_did("wide", 0);
                break;
        case v_narrow :
                pangle = ec_did("narrow", 0);
                break;
        default :
                pangle = ec_did("unknown", 0);
        }
        
        switch (vquality) {
        case v_high :
                pquality = ec_did("high", 0);
                break;
        case v_low :
                pquality = ec_did("low", 0);
                break;
        default :
                pquality = ec_did("unknown", 0);
        }

        ec_ok(ec_unify_arg(1, ec_atom(pangle)), &res);
        ec_ok(ec_unify_arg(2, ec_atom(pquality)), &res);
        ec_ok(ec_unify_arg(3, ec_double(stamina)), &res);
        ec_ok(ec_unify_arg(4, ec_double(effort)), &res);
        ec_ok(ec_unify_arg(5, ec_double(speed)), &res);
        ec_ok(ec_unify_arg(6, ec_long(kicks)), &res);
        ec_ok(ec_unify_arg(7, ec_long(dashs)), &res);
        ec_ok(ec_unify_arg(8, ec_long(turns)), &res);
        ec_ok(ec_unify_arg(9, ec_long(says)), &res);

        return res;
}

/**
 * players stamina
 *
 * Prolog: stamina(?Stamina)
 */
int p_stamina()
{
        return ec_unify_arg(1, ec_double(RSC_stamina()));
}

/**
 * players effort
 *
 * Prolog: effort(?Effort)
 */
int p_effort()
{
        return ec_unify_arg(1, ec_double(RSC_effort()));
}

/**
 * players recovery value
 *
 * Prolog: recovery(?Recovery)
 */
int p_recovery()
{
        return ec_unify_arg(1, ec_double(RSC_recovery()));
}

/**
 * players speed
 *
 * Prolog: speed(?Speed)
 */
int p_speed()
{
        return ec_unify_arg(1, ec_double(RSC_speed()));
}

/**
 * number of dashs
 *
 * dashs(+N)
 */
int p_dashs()
{
        return ec_unify_arg(1, ec_long(RSC_dashs()));
}

/**
 * number of turns
 *
 * turns(+N)
 */
int p_turns()
{
        return ec_unify_arg(1, ec_long(RSC_turns()));
}
 
/**
 * number of turn_necks
 *
 * turn_necks(+N)
 */
int p_turn_necks()
{
        return ec_unify_arg(1, ec_long(RSC_turn_necks()));
}
 
/**
 * number of says
 *
 * says(+N)
 */
int p_says()
{
        return ec_unify_arg(1, ec_long(RSC_says()));
}

/**
 * number of goals
 *
 * goals(+Time, +Side, ?N)
 */
int p_goals()
{
        int     res = PSUCCEED;

        pword   arg1 = ec_arg(1);
        pword   arg2 = ec_arg(2);
        
        dident  pside;
        char    *side;
        int     goals;
        long int time;

        ec_ok(ec_get_long(arg1, &time), &res);
        ec_ok(ec_get_atom(arg2, &pside), &res);
        side = ec_did_name(pside);
        
        switch (side[0]) {
        case 'l' :
                goals = RSC_goals(time, id_goal_left);
                break;
        case 'r' :
                goals = RSC_goals(time, id_goal_right);
                break;
        default :
                ec_err(RANGE_ERROR, &res);
                goals = 0;
        }

        
        ec_ok(ec_unify_arg(3, ec_long(goals)), &res);

        return res;
}
 
/**
 * moves client to a position on the field before kick off
 *
 * Prolog: moveto(+X, +Y)
 */
int p_moveto()
{
        int res = PSUCCEED;
        
        double x;
        double y;

        if (ec_ok(ec_get_number(ec_arg(1), &x), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &y), &res)) {
                RSC_moveto(x, y);
        }
        return res;
}

/**
 * turns client around. positive values mean turning clockwise.
 *
 * Prolog: turn(+Angle)
 */
int p_turn()
{
        int    res = PSUCCEED;        
        double angle;

        if (ec_ok(ec_get_number(ec_arg(1), &angle), &res)) {
                RSC_turn(angle);
        }
        return res;
}

/**
 * turns client's neck. value is added to current neck angle.
 *
 * Prolog: turn(+Angled)
 */
int p_turn_neck()
{
        int    res = PSUCCEED;        
        double angled;

        if (ec_ok(ec_get_number(ec_arg(1), &angled), &res)) {
                RSC_turn_neck(angled);
        }
        return res;
}

/**
 * accelerates the client. -30 <= dash power <= 100
 *
 * Prolog: dash(+power)
 */
int p_dash()
{
        int    res = PSUCCEED;        
        double power;

        if (ec_ok(ec_get_number(ec_arg(1), &power), &res)) {        
                RSC_dash(power);
        }
        return res;
}

/**
 * kicks against the ball.
 *
 * Prolog: kick(+Power, +Angle)
 */
int p_kick()
{
        int     res = PSUCCEED;
        double  power;
        double  angle;
                
        if (ec_ok(ec_get_number(ec_arg(1), &power), &res) &&
            ec_ok(ec_get_number(ec_arg(2), &angle), &res)) {
                RSC_kick(power, angle);
        }
        return res;
}

/**
 * shout text across the field.
 *
 * Prolog: say(+Text)
 */
int p_say()
{
        int     res = PSUCCEED;
        char    *txt;
                
        if (ec_ok(ec_get_string(ec_arg(1), &txt), &res)) {
                RSC_say(txt);
        }
        return res;
}

/**
 * goalie catch command
 *
 * Prolog: catch(+Angle)
 */
int p_catch()
{
        int    res = PSUCCEED;        
        double angle;

        if (ec_ok(ec_get_number(ec_arg(1), &angle), &res)) {
                RSC_catch(angle);
        }
        return res;
}

/**
 * kicks the ball a certain distance.
 *
 * Prolog: kick2(+Dist, +Dir)
 */
int p_kick2() 
{
        double dir;
        double dist;
        
        int     res = PSUCCEED;

        if (ec_ok(ec_get_number(ec_arg(1), &dist), &res)
            && ec_ok(ec_get_number(ec_arg(2), &dir), &res)) {
                RSC_kick2(dist, dir);
        }

        return res;
}

/**
 * kick whereever you want to kick to - experimental and still not working
 *
 * Prolog: kick3(+Dist, +Dir, +Wise)
 */
int p_kick3() 
{
        double dir;
        double dist;
        long int wise;
        
        int     res = PSUCCEED;

        if (ec_ok(ec_get_number(ec_arg(1), &dist), &res)
            && ec_ok(ec_get_number(ec_arg(2), &dir), &res)
            && ec_ok(ec_get_long(ec_arg(3), &wise), &res)) {
                RSC_kick3(dist, dir, wise);
        }        
        return res;
}


/**
 * gets ball position
 *
 * Prolog: b_ball(?Id, pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_ball_dep()
{
        long int i;
        int res = PSUCCEED;
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(3), &i), &res)) {
                RSC_ball_dep(&od);

                ec_ok(ec_unify_arg(1, ec_long(0)), &res);
                ec_ok(ec_unify_arg(2, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }
        
        return res;
}

/**
 * gets goal positions
 *
 * Prolog: b_goal(?Side, pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_goal_dep() 
{
        long int i;
        int     res = PSUCCEED;
        object_data_t od;
        dident  pside;

        if (ec_ok(ec_get_long(ec_arg(3), &i), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(3, next.val, next.tag);
                i--;

                if (i < 2) {
                        switch (i) {
                        case id_goal_left :
                                RSC_goal_dep(id_goal_left, &od);
                                pside = ec_did("l", 0);
                                break;
                        case id_goal_right :
                                RSC_goal_dep(id_goal_right, &od);
                                pside = ec_did("r", 0);
                        }
                        res = ec_unify_arg(1, ec_atom(pside));
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(2, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns positions of partner with index number 'index'
 * (0 <= index < Number of team members seen)
 * fails if index >= Number of team members seen
 *
 * Prolog: partner (?Id,
 *                  pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_partner_dep()
{
        long int i;
        int   res = PSUCCEED;
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(3), &i), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(3, next.val, next.tag);
                i--;
                
                if (i < RSC_n_partner_dep()) {
                        RSC_partner_dep(i, &od);
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(1, ec_long(od.id)), &res);
                ec_ok(ec_unify_arg(2, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns positions of opponent with index number 'index'
 * (0 <= index < Number of opponents seen)
 * fails if index >= Number of opponents seen
 *
 * Prolog: b_opponent (?Id,
 *                     pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_opponent_dep()
{
        long int i;
        int   res = PSUCCEED;
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(3), &i), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(3, next.val, next.tag);
                i--;
                
                if (i < RSC_n_opponent_dep()) {
                        RSC_opponent_dep(i, &od);
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(1, ec_long(od.id)), &res);
                ec_ok(ec_unify_arg(2, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns positions of unidentified player with index number 'index'
 * (0 <= index < Number of unidentified players seen)
 * fails if index >= Number of unidentified players seen
 *
 * Prolog: b_player (?Id,
 *            pos(?Dist, ?Dir, ?DDist, ?DDir, ?X, ?Y, ?Bodydir, ?Facedir), +I)
 */
int p_player_dep()
{
        long int i;
        int   res = PSUCCEED;
        object_data_t od;

        if (ec_ok(ec_get_long(ec_arg(3), &i), &res)) {
                pword next = ec_long(i + 1);
                res = ec_remember(3, next.val, next.tag);
                i--;
                
                if (i < RSC_n_player_dep()) {
                        RSC_player_dep(i, &od);
                } else {
                        cut_external();
                        res = PFAIL;
                }

                ec_ok(ec_unify_arg(1, ec_long(od.id)), &res);
                ec_ok(ec_unify_arg(2, ec_term(ec_did("pos", 8),
                                              ec_double(od.dist),
                                              ec_double(od.dir),
                                              ec_double(od.ddist),
                                              ec_double(od.ddir),
                                              ec_double(od.x),
                                              ec_double(od.y),
                                              ec_double(od.bodydir),
                                              ec_double(od.facedir))), &res);
        }

        return res;
}

/**
 * returns/sets position
 *
 * Prolog self(?X, ?Y, ?Dir)
 */
int p_self_dep()
{
        pword arg1 = ec_arg(1);
        pword arg2 = ec_arg(2);
        pword arg3 = ec_arg(3);

        int res = PSUCCEED;
        
        double xpos, ypos, angle;
        
        if (ec_ok(ec_get_number(arg1, &xpos), &res)
            && ec_ok(ec_get_number(arg2, &ypos), &res)
            && ec_ok(ec_get_number(arg3, &angle), &res)) {
                RSC_self_dep(&xpos, &ypos, &angle, SET);
        } else {
                res = PSUCCEED;
                RSC_self_dep(&xpos, &ypos, &angle, GET);
                arg1 = ec_double(xpos);
                arg2 = ec_double(ypos);
                arg3 = ec_double(angle);
                ec_ok(ec_unify_arg(1, arg1), &res);
                ec_ok(ec_unify_arg(2, arg2), &res);
                ec_ok(ec_unify_arg(3, arg3), &res);
        }
        
        return res;
}
