/* -*- mode: c++ -*-
   this file is part of the robolog libraries
   fifo implementation
   Copyright (C) Oct 29 1997 Oliver Obst
   $Id: fifo.C,v 1.3 1999/02/03 08:15:05 frvit Exp $

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>
#include <values.h>
// #if defined DEBUG_FIFO || defined DEBUG_MEMORY
#include <iostream.h>
// #endif

#include "fifo.h"

// --- public constructor -----------------------------------------------------
// initialize a fifo. if max==0, the number of fifo elements is unlimited.
Fifo::Fifo(int max)
{
        head    = 0;
        tail    = 0;
        fsize   = 0;
        maxfsize = max;
}


// --- public destructor ------------------------------------------------------
Fifo :: ~Fifo()
{
        clear();
}


// --- public -----------------------------------------------------------------
// read the first element
//      char *data              pointer to memory area
//      bool *move              movement flag
//      bool lock               true if locking needed (use false for irq read)
// return int                   bytes read
int Fifo::read(char *data, bool *move, bool locking)
{
#ifdef DEBUG_FIFO
        cerr << "FIFO: read(char *data)\n";
#endif
        int l = 0;
        
        if (locking) lock();
        
        if (fsize > 0) {
                l = head->size;
                memcpy(data, head->atom, l);
                if (move)
                        memcpy(move, &(head->movement), sizeof(bool));
                _del_top();
        }
        
        if (locking) unlock();
        return l;
}


// --- public -----------------------------------------------------------------
// put a string onto the fifo
//      char *text              text string to store
//      bool move              movement flag
// return true if success
bool Fifo::write(char *text, bool move)
{
#ifdef DEBUG_FIFO
        cerr << "FIFO: write(char *text)\n";
#endif

        bool b = false;
        
        while (maxfsize > 0 && fsize >= maxfsize);
        lock();
                
        if (b = (maxfsize == 0 || fsize < maxfsize)) {
                Fcell *last = new Fcell;
#if defined DEBUG_MEMORY || defined DEBUG_FIFO
                cerr << "fifo.C: write(char *, bool): "
                     << "last = new Fcell\n";
#endif
                int l = strlen(text) + 1;
                last->size = l;
                last->next = tail;
                last->prev = 0;
                last->atom = new char[ l ];
#if defined DEBUG_MEMORY || defined DEBUG_FIFO
                cerr << "fifo.C: write(char *, bool): "
                     << "last->atom = new char["
                     << l << "]\n";
#endif                   
                memcpy(last->atom, text, l);
                memcpy(&(last->movement), &move, sizeof(move));
                if (tail != 0)
                        tail->prev = last;
                else
                        head = last;
                tail = last;
                fsize++;
        }
        
        unlock();
        return b;
}


// --- public -----------------------------------------------------------------
// put ``size'' bytes of data onto the fifo
//      char *data              data to store
//      int  size               number of bytes
//      bool move              movement flag
// return true if success
bool Fifo::write(char *data, int size, bool move)
{
#ifdef DEBUG_FIFO
        cerr << "FIFO: write(char *data, int size)\n";
#endif
        bool b = false;

        while (maxfsize > 0 && fsize >= maxfsize);
        lock();

        if ((b = (maxfsize == 0 || fsize < maxfsize)) && size > 0) {
                Fcell *last = new Fcell;
#ifdef DEBUG_MEMORY
                cerr << "fifo.C: write(char *, int, bool): "
                     << "last = new Fcell\n";
#endif                
                last->size = size;
                last->next = tail;
                last->prev = 0;
                last->atom = new char[ size ];
#ifdef DEBUG_MEMORY
                cerr << "fifo.C: write(char *, bool): "
                     << "last->atom = new char["
                     << size << "]\n";
#endif                
                memcpy(last->atom, data, size);
                memcpy(&(last->movement), &move, sizeof(move));
                if ( tail != 0 )
                        tail->prev = last;
                else
                        head = last;
                tail = last;
                fsize++;
        }
        
        unlock();
        return b;
}


// --- public -----------------------------------------------------------------
// look if the top element is a movement command
bool Fifo::peek_move()
{
        bool b = false;
        if (fsize > 0) b = head->movement;
        return b;
}


// --- public -----------------------------------------------------------------
// delete the head element
void Fifo::del_top()
{
        lock();
        _del_top();
        unlock();
}


// --- public -----------------------------------------------------------------
// empty the whole fifo
void Fifo::clear() 
{
        lock();
        while (fsize > 0)
                _del_top();
        unlock();
}


// --- public -----------------------------------------------------------------
// return true if fifo is empty
bool Fifo::empty()
{
        return (fsize == 0);
}


// --- public -----------------------------------------------------------------
// return true if maximum amount of fifo elements is used
bool Fifo::full()
{
        return (maxfsize > 0 && fsize == maxfsize);
}


// --- public -----------------------------------------------------------------
// return maximum number of elements
int Fifo :: size()
{
        return fsize;
}


// --- protected --------------------------------------------------------------
// lock fifo by blocking timer signals
void Fifo::lock() 
{
        sigset_t mask;
        
        sigemptyset(&mask);
        sigaddset(&mask, SIGALRM);
        
        sigprocmask(SIG_BLOCK, &mask, 0);
}

// --- protected --------------------------------------------------------------
// unlock fifo by unblocking timer signals
void Fifo::unlock() 
{
        sigset_t mask;
        
        sigemptyset(&mask);
        sigaddset(&mask, SIGALRM);
        
        sigprocmask(SIG_UNBLOCK, &mask, 0);
}

// --- protected --------------------------------------------------------------
// delete top fifo element
void Fifo :: _del_top()
{
#ifdef DEBUG_FIFO
        cerr << "FIFO: _del_top()\n";
#endif        
        if (fsize > 0) {
                Fcell *second = head->prev;
                delete head->atom;
                delete head;
#ifdef DEBUG_MEMORY
                cerr << "fifo.C: _del_top(): delete head->atom\n";
                cerr << "fifo.C: _del_top(): delete head\n";
#endif                
                head = second;
                fsize--;
                if ( head !=0 ) 
                        head->next = 0;
                else
                        tail = 0;
        }
}

