/* -*- mode: c++ -*-
   
   this file is part of the robolog libraries
   Copyright (C) Oct 29 1997 Oliver Obst
   a fifo definition
   $Id: fifo.h,v 1.4 1999/05/05 13:56:07 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef FIFO_H
#define FIFO_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>

/**
 * This fifo can store elements containing a command (a string),
 * the size of the string and a movement flag (indicates commands that
 * move the client).
 * Additionally, the elements contain pointers to the previous and
 * to the next element.
 *
 * The static design of this class was chosen for reasons of efficience.
 *
 * @author      Oliver Obst <frvit@uni-koblenz.de>
 * @version     $Id: fifo.h,v 1.4 1999/05/05 13:56:07 frvit Exp $
 * @short       A fifo (stack) for soccer client/server commands.
 */
class Fifo
{
public:
        /**
         * Creates a fifo (stack) of size <code>max</code>.
         *
         * @param max           size of the stack; if <code>max == 0</code>
         *                      the size of the stack is unlimited (size of
         *                       memory)
         */
        Fifo(int max = 5);
        /**
         * Destructor. Deletes all fifo elements first.
         */
        ~Fifo();

        /**
         * Reads the topmost element and deletes it from the stack.
         * Some (poor) locking is done while accessing the fifo
         * to prevent multiple access at a time.
         * @param data          buffer to store the data into
         * @param move          will be set to <code>true</code> if
         *                      buffer contains a movement command
         * @return              number of bytes read
         */
        int     read(char *data, bool *move=0, bool locking=true);
        bool    write(char *text, bool move=false);
        bool    write(char *data, int size, bool move=false);
        bool    peek_move();
        void    del_top();
        void    clear();
        bool    empty();
        bool    full();
        int     size();        

protected:
        struct Fcell
        {
                char    *atom;
                int     size;
                bool    movement;
                Fcell   *next;
                Fcell   *prev;
        };

        void    _del_top();

        void    lock();
        void    unlock();

        Fcell   *head;
        Fcell   *tail;
        volatile int     fsize;
        int     maxfsize;
};
#endif /* FIFO_H */
