/* -*- mode: c++ -*-
   this file is part of the robolog libraries
   mathematical stuff implementations
   Copyright (c) Mar 22 1998 Oliver Obst
   $Id: maths.C,v 1.4 1999/05/05 14:08:38 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* parts of the code in this file are taken from
 * CMUnited98 (soccer client for Robocup98)
 * Peter Stone <pstone@cs.cmu.edu>
 * Computer Science Department
 * Carnegie Mellon University
 * Copyright (C) 1998 Peter Stone
 *
 * CMUnited-98 was created by Peter Stone, Manuela Veloso, and Patrick Riley
 *
 * You may copy and distribute this program freely as long as you retain
 * this notice.
 * If you make any changes or have any comments we would appreciate a message.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include "constants.h"
#include "debug.h"
#include "maths.h"

// ----------------------------------------------------------------------------
// compute e^ix
Cvector iexp(double alpha, bool deg)
{
        if (deg) alpha = deg2rad(alpha);
        return Cvector(cos(alpha), sin(alpha));
}

// ----------------------------------------------------------------------------
// translate deg into rad angle
double deg2rad(double deg)
{
        return deg * m_deg2rad;
}

// ----------------------------------------------------------------------------
// translate rad into deg angle
double rad2deg(double rad)
{
        return rad * m_rad2deg;
}

// ----------------------------------------------------------------------------
// normalize value x to 0 if x < -9000
double normalize_value(double x)
{
        return (x < - 9000 ? 0 : x);
}

// ----------------------------------------------------------------------------
// normalize angle x to -180 deg. < x <= 180 deg.
//             resp. to -PI rad. < x <= PI rad.
double normalize_angle(double x, bool deg)
{
        if (deg) {
                while (x <= -180.0) x += 360.0;
                while (x > 180.0) x-= 360.0;
        } else {
                while (x <= -M_PI) x += M_PI + M_PI;
                while (x > M_PI) x -= (M_PI + M_PI);
        }

        return x;
}

// ----------------------------------------------------------------------------
// normalize angle x to -180 deg. < x <= 180 deg.
AngleDeg normalize_angle_deg(AngleDeg x)
{
        while (x <= -180.0) x += 360.0;
        while (x > 180.0) x-= 360.0;
        return x;
}

// ----------------------------------------------------------------------------
// normalize angle x to -PI rad. < x <= PI rad.
AngleRad normalize_angle_rad(AngleRad x)
{
        while (x <= -M_PI) x += M_PI + M_PI;
        while (x > M_PI) x -= (M_PI + M_PI);
        return x;
}

// ----------------------------------------------------------------------------
int int_random(int n)
{
        static int FirstTime = true;
  
        if (FirstTime) {
                /* initialize the random number seed. */
                timeval tp;
                gettimeofday(&tp, 0);
                srandom((unsigned int) tp.tv_usec);
                FirstTime = false;
        }

        if (n > 2)
                return(random() % n);
        else if (n == 2)
                return(((random() % 112) >= 56 ) ? 0 : 1);
        else if (n == 1)
                return(0);
        else {
                ROBOBUG(ROBOBUG_ERROR,MATHS_RANDOM,
                        "You called int_random(<=0)");
                return(0);
        }
}

// ----------------------------------------------------------------------------
double range_random(double lo, double hi)
{
        int x1 = int_random(10000);
        int x2 = int_random(10000);
        double r = (((double) x1) +
                    10000.0 * ((double) x2))/(10000.0 * 10000.0);
        return (lo + (hi - lo) * r);
}

// AngleRad -------------------------------------------------------------------
// --- public constructor -----------------------------------------------------
AngleRad::AngleRad(double d)
{
        a = d;
        _normalize();
}

// --- public copy constructor ------------------------------------------------
AngleRad::AngleRad(const AngleRad &ang)
{
        a = ang.a;
}

// --- public conversion constructor ------------------------------------------
AngleRad::AngleRad(const AngleDeg &ang)
{
        a = ang.rad();
}

// --- public -----------------------------------------------------------------
// unary -
AngleRad AngleRad::operator -()
{
        a = -a;
        return *this;
}

// --- public -----------------------------------------------------------------
// addition
AngleRad AngleRad::operator +(const AngleRad &ang)
{
        a += ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
// subtraction
AngleRad AngleRad::operator -(const AngleRad &ang)
{
        a -= ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
// assign, no range check
AngleRad AngleRad::operator =(const AngleRad &ang)
{
        a = ang.a;
        return *this;
}

// --- public -----------------------------------------------------------------
// assign, no range check
AngleRad AngleRad::operator =(const AngleDeg &ang)
{
        a = ang.rad();
        return *this;
}

// --- public -----------------------------------------------------------------
// assign, range check
AngleRad AngleRad::operator =(const double &d)
{
        a = d;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleRad AngleRad::operator +=(const AngleRad &ang)
{
        a += ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleRad AngleRad::operator -=(const AngleRad &ang)
{
        a -= ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleRad AngleRad::operator *=(const double &d)
{
        a *= d;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleRad AngleRad::operator /=(const double &d)
{
        a /= d;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
int AngleRad::operator ==(const AngleRad& ang)
{
        return (a == ang.a);
}

// --- public -----------------------------------------------------------------
int AngleRad::operator !=(const AngleRad& ang)
{
        return (a != ang.a);
}

// --- protected --------------------------------------------------------------
void AngleRad::_normalize()
{
        while (a <= -M_PI) a += M_PI + M_PI;
        while (a > M_PI) a -= (M_PI + M_PI);
}

// AngleDeg -------------------------------------------------------------------

// --- public constructor -----------------------------------------------------
AngleDeg::AngleDeg(double d)
{
        a = d;
        _normalize();
}

// --- public copy constructor ------------------------------------------------
AngleDeg::AngleDeg(const AngleDeg &ang)
{
        a = ang.a;
}

// --- public conversion constructor ------------------------------------------
AngleDeg::AngleDeg(const AngleRad &ang)
{
        a = ang.deg();
}

// --- public -----------------------------------------------------------------
// unary -
AngleDeg AngleDeg::operator -()
{
        a = -a;
        return *this;
}

// --- public -----------------------------------------------------------------
// addition
AngleDeg AngleDeg::operator +(const AngleDeg &ang)
{
        a += ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
// subtraction
AngleDeg AngleDeg::operator -(const AngleDeg &ang)
{
        a -= ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
// assign, no range check
AngleDeg AngleDeg::operator =(const AngleDeg &ang)
{
        a = ang.a;
        return *this;
}

// --- public -----------------------------------------------------------------
// assign, no range check
AngleDeg AngleDeg::operator =(const AngleRad &ang)
{
        a = ang.deg();
        return *this;
}

// --- public -----------------------------------------------------------------
// assign, range check
AngleDeg AngleDeg::operator =(const double &d)
{
        a = d;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleDeg AngleDeg::operator +=(const AngleDeg &ang)
{
        a += ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleDeg AngleDeg::operator -=(const AngleDeg &ang)
{
        a -= ang.a;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleDeg AngleDeg::operator *=(const double &d)
{
        a *= d;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
AngleDeg AngleDeg::operator /=(const double &d)
{
        a /= d;
        _normalize();
        return *this;
}

// --- public -----------------------------------------------------------------
int AngleDeg::operator ==(const AngleDeg& ang)
{
        return (a == ang.a);
}

// --- public -----------------------------------------------------------------
int AngleDeg::operator !=(const AngleDeg& ang)
{
        return (a != ang.a);
}

// --- protected --------------------------------------------------------------
void AngleDeg::_normalize()
{
        while (a <= -180.0) a += 360.0;
        while (a > 180.0) a-= 360.0;
}


// Cvector --------------------------------------------------------------------

// --- public constructor -----------------------------------------------------
Cvector::Cvector(double vx, double vy)
{
        myx = vx;
        myy = vy;
}

// --- public copy constructor ------------------------------------------------
Cvector::Cvector(const Cvector &v)
{
        myx = v.myx;
        myy = v.myy;
}

// --- public conversion constructor ------------------------------------------
Cvector::Cvector(double dist, AngleDeg dir)
{
        myx = dist * Cos(dir);
        myy = dist * Sin(dir);
}

// --- public conversion constructor ------------------------------------------
Cvector::Cvector(double dist, AngleRad dir)
{
        myx = dist * cos(dir);
        myy = dist * sin(dir);        
}

// --- public -----------------------------------------------------------------
// set Cvector to (vx, vy)
Cvector& Cvector::set(double vx, double vy)
{
        myx = vx;
        myy = vy;
        return *this;
}

// --- public -----------------------------------------------------------------
// set Cvector to polar (dist, dir)
Cvector& Cvector::set(double dist, AngleDeg dir)
{
        myx = dist * Cos(dir);
        myy = dist * Sin(dir);
        return *this;
}

// --- public -----------------------------------------------------------------
// set Cvector to polar (dist, dir)
Cvector& Cvector::set(double dist, AngleRad dir)
{
        myx = dist * cos(dir);
        myy = dist * sin(dir);
        return *this;
}

// --- public -----------------------------------------------------------------
// set x coordinate
inline void Cvector::setx(double vx)
{
        myx = vx;
}

// --- public -----------------------------------------------------------------
// set y coordinate
inline void Cvector::sety(double vy)
{
        myy = vy;
}

// --- public -----------------------------------------------------------------
inline double Cvector::x() const
{
        return myx;
}

// --- public -----------------------------------------------------------------
inline double Cvector::y() const
{
        return myy;
}

// --- public -----------------------------------------------------------------
// calculate distance from (0, 0) to (x, y)
double Cvector::len() const
{
        return sqrt(myx*myx + myy*myy);
}

// --- public -----------------------------------------------------------------
// calculate distance from (0, 0) to (x, y)   (faster than len())
double Cvector::sqrlen() const
{
        return myx*myx + myy*myy;
}

// --- public -----------------------------------------------------------------
// distance from a point (x1, y1) to (x, y)
inline double Cvector::dist(const Cvector &v1) const
{
        return (*this-v1).len();
}

// --- public -----------------------------------------------------------------
// square distance from a point (x1, y1) to (x, y)
inline double Cvector::sqrdist(const Cvector &v1) const
{
        return (*this-v1).sqrlen();
}

// --- public -----------------------------------------------------------------
// calculate angle (RAD) of Cvector. 
double Cvector::dir(bool deg) const
{
        double d;
        d = (myx == 0 && myy == 0) ? 0 : atan2(myy, myx);
        while (d > M_PI) d -= (M_PI + M_PI);
        while (d < -M_PI) d += M_PI + M_PI;
        if (deg) d = rad2deg(d);
        return d;
}

// --- public -----------------------------------------------------------------
// rotate vector angle degr. or rad., if deg==true resp. rad==false
void Cvector::rotate(double angle, bool deg)
{
        double d=dir();
        double l=len();

        if (deg) angle = deg2rad(angle);
        
        myx = l * cos(d + angle);
        myy = l * sin(d + angle);
}

// --- public -----------------------------------------------------------------
// return a vector which is *this rotated for deg degrees
Cvector Cvector::rotate(AngleDeg ang) const
{
        return Cvector(this->len()*Cos(this->dir(true) + ang),
                       this->len()*Sin(this->dir(true) + ang));
}
                
// --- public -----------------------------------------------------------------
// return a vector which is *this rotated for rad degrees
Cvector Cvector::rotate(AngleRad ang) const
{
        return Cvector(this->len()*cos(this->dir(false) + ang),
                       this->len()*sin(this->dir(false) + ang));
}

// --- public -----------------------------------------------------------------
Cvector Cvector::global2relative(Cvector orig, AngleDeg ang)
{
        return (*this - orig).rotate(-ang);
}

// --- public -----------------------------------------------------------------
Cvector Cvector::global2relative(Cvector orig, AngleRad ang)
{
        return (*this - orig).rotate(-ang);        
}

// --- public -----------------------------------------------------------------
Cvector Cvector::relative2global(Cvector orig, AngleDeg ang)
{
        return (rotate(ang) + orig);
}

// --- public -----------------------------------------------------------------
Cvector Cvector::relative2global(Cvector orig, AngleRad ang)
{
        return (rotate(ang) + orig);
}

// --- public operator --------------------------------------------------------
// -(x, y) = (-x, -y)
Cvector Cvector::operator -() const
{
        return Cvector(-myx, -myy);
}

// --- public operator --------------------------------------------------------
// (x1, y1) + (x2, y2) = (x1+x2, y1+y2)
Cvector Cvector::operator +(const Cvector &v) const
{
        return Cvector(myx + v.myx, myy + v.myy);
}


// --- public operator --------------------------------------------------------
// (x1, y1) - (x2, y2) = (x1-x2, y1-y2)
Cvector Cvector::operator -(const Cvector &v) const
{
        return Cvector(myx - v.myx, myy - v.myy);
}

// --- public operator --------------------------------------------------------
// complex number multiplication
// (x1, y1) * (x2, y2) = (x1*x2-y1*y2, x1*y2+y1*x2) 
Cvector Cvector::operator *(const Cvector &v) const
{
        return Cvector(myx*v.myx - myy*v.myy, myx*v.myy + myy*v.myx);
}

// --- public operator --------------------------------------------------------
// (x, y) * d = (x*d, y*d)
Cvector Cvector::operator *(const double &d) const
{
        return Cvector(myx*d, myy*d);
}


// --- public operator --------------------------------------------------------
// complex number division
// (x1, y1) / (x2, y2) = ((x1*x2+y1*y2) / (x2*x2 + y2*y2),
//                        (x2*y1 - y2*x1) / (x2*x2 + y2*y2))
Cvector Cvector::operator /(const Cvector &v) const
{
        return Cvector((myx*v.myx+myy*v.myy) / (v.myx*v.myx + v.myy*v.myy),
                       (v.myx*myy - v.myy*myx) / (v.myx*v.myx + v.myy*v.myy));
}

// --- public operator --------------------------------------------------------
// (x, y) / d = (x/d, y/d)
Cvector Cvector::operator /(const double &d) const
{
        return Cvector(myx/d, myy/d);
}


// --- public operator --------------------------------------------------------
// (vx, vy)
Cvector& Cvector::operator =(const Cvector &v)
{
        myx = v.myx;
        myy = v.myy;
        return *this;
}


// --- public operator --------------------------------------------------------
// (d, d)
Cvector& Cvector::operator =(const double &d)
{
        myx = myy = d;
        return *this;
}


// --- public operator --------------------------------------------------------
// (x, y) = (x, y) + (vx, vy)
Cvector& Cvector::operator +=(const Cvector &v)
{
        myx += v.myx;
        myy += v.myy;
        return *this;
}


// --- public operator --------------------------------------------------------
// (x, y) = (x, y) + d                  == (x+d, y+d)
Cvector& Cvector::operator +=(const double &d)
{
        myx += d;
        myy += d;
        return *this;
}


// --- public operator --------------------------------------------------------
// (x, y) = (x, y) - (vx, vy)           == (x-vx, y-vy)
Cvector& Cvector::operator -=(const Cvector &v)
{
        myx -= v.myx;
        myy -= v.myy;
        return *this;
}


// --- public operator --------------------------------------------------------
// (x, y) = (x, y) - d                  == (x-d, y-d)
Cvector& Cvector::operator -=(const double &d)
{
        myx -= d;
        myy -= d;
        return *this;
}


// --- public operator --------------------------------------------------------
// (x, y) = (x, y) * d
Cvector& Cvector::operator *=(const double &d)
{
        myx *= d;
        myy *= d;
        return *this;
}


// --- friend operator --------------------------------------------------------
// (x1, y1) = d * (x2, y2)              == (d*x2, d*y2)
Cvector operator *(const double &d, const Cvector &v)
{
        return Cvector(d*v.myx, d*v.myy);
}

        
// --- friend operator --------------------------------------------------------
// i:= (x1, y1) == (x2, y2)
int operator ==(const Cvector &v, const Cvector &w)
{
        return (v.myx == w.myx && v.myy == w.myy);
}


// --- friend operator --------------------------------------------------------
// i:= (x1, y1) == d
int operator ==(const Cvector &v, const double &d)
{
        return (v.myx == d && v.myy == d);
}


// --- friend operator --------------------------------------------------------
// i:= (x1, y1) != (x2, y2)
int operator !=(const Cvector &v, const Cvector &w)
{
        return (v.myx != w.myx || v.myy != w.myy);
}


// --- friend operator --------------------------------------------------------
// i:= (x1, y1) != d
int operator !=(const Cvector &v, const double &d)
{
        return (v.myx != d || v.myy != d);
}

#if 0
// --- friend operator --------------------------------------------------------
// send "(x, y)" to output stream
ostream& operator <<(ostream &os, const Cvector &v)
{
        return os << "(" << v.myx << ", " << v.myy << ")";
}
#endif

// --- public constructor -----------------------------------------------------
Pvector::Pvector(double dist, double dir, bool deg)
{
        set(dist, dir, deg);
}


// --- public destructor ------------------------------------------------------
Pvector::~Pvector()
{
}


// --- public -----------------------------------------------------------------
Pvector& Pvector::set(double dist, double dir, bool deg)
{
        mydist = dist;

        deg = deg || (dir == bearing_unknown);
        
        if (deg) {
                dir_deg = dir;
                deg_valid = true;
                if (rad_valid = (dir == bearing_unknown))
                        dir_rad = bearing_unknown;
        } else {
                dir_rad = dir;
                deg_valid = false;
                rad_valid = true;
        }
        return *this;
}


// --- public -----------------------------------------------------------------
Pvector& Pvector::set_cart(double vx, double vy)
{
        rad_valid = true;
        deg_valid = false;
        mydist = sqrt(vx*vx + vy*vy);
        dir_rad = atan2(vy, vx);
        return *this;
}


// --- public -----------------------------------------------------------------
void Pvector::set_dist(const double dist)
{
        mydist = dist;
}


// --- public -----------------------------------------------------------------
void Pvector::set_dir(const double dir, bool deg)
{
        deg_valid = deg || dir == bearing_unknown;
        rad_valid = !deg || dir == bearing_unknown;
        
        if (deg_valid) {
                dir_deg = dir;
                if (rad_valid) dir_rad = bearing_unknown;
        } else {
                dir_rad = dir;
                
        }
}


// --- public -----------------------------------------------------------------
double Pvector::dist() const
{
        return mydist;
}


// --- public -----------------------------------------------------------------
double Pvector::dir(bool deg)
{
        if (deg && !deg_valid) {
                deg_valid = true;
                dir_deg = rad2deg(dir_rad);
        }
        if (!deg && !rad_valid) {
                rad_valid = true;
                dir_rad = deg2rad(dir_deg);
        }
        
        return (deg ? dir_deg : dir_rad);
}


// --- public -----------------------------------------------------------------
Pvector& Pvector::operator =(const Pvector &v)
{
        mydist  = v.mydist;
        dir_deg = v.dir_deg;
        dir_rad = v.dir_rad;
        deg_valid = v.deg_valid;
        rad_valid = v.rad_valid;
        return *this;
}

                

// --- friend operator --------------------------------------------------------
int operator ==(Pvector &v0, Pvector &v1)
{
        return (v0.mydist == v1.mydist && v0.dir() == v1.dir());
}


// --- public constructor -----------------------------------------------------
Xvector::Xvector(double vx, double vy, bool auth) : Cvector(vx, vy)
{
        valid = auth;
}


// --- public destructor ------------------------------------------------------
Xvector::~Xvector()
{
}


// --- public -----------------------------------------------------------------
void Xvector::set_authentic(bool auth)
{
        valid = auth;
}


// --- public -----------------------------------------------------------------
bool Xvector::authentic() const
{
        return valid;
}

// ----------------------------------------------------------------------------
// Line -----------------------------------------------------------------------
// most of this is CMU code

// --- public constructor -----------------------------------------------------
Line::Line(double x_coef, double y_coef, double constant) 
{
        A = x_coef;
        B = y_coef;
        C = constant;
}

// --- public -----------------------------------------------------------------
void Line::LineFromTwoPoints(Cvector pt1, Cvector pt2)
{
        double temp = (pt2.x() - pt1.x());
        if (fabs(temp) < DOUBLE_EPS) {
                if (fabs(pt2.y() - pt1.y()) < DOUBLE_EPS) {
                        ROBOBUG(ROBOBUG_ERROR,MATHS_LINE,
                                "LineFromTwoPoints: points are equal");
                }
                
                A = 1;
                B = 0;
        } else {
                double m = (pt2.y() - pt1.y()) / temp;
                A = -m;
                B = 1;
        }  
        C = -(A * pt2.x() + B * pt2.y());
}

// -- public ------------------------------------------------------------------
void Line::LineFromRay(Cvector orig, Cvector dir)   
{
        if (fabs(dir.y()) < DOUBLE_EPS && fabs(dir.x()) < DOUBLE_EPS)
                ROBOBUG(ROBOBUG_ERROR,MATHS_LINE,
                        "LineFromRay: dir is zero");
        LineFromTwoPoints(orig, orig + dir);
}

// --- public -----------------------------------------------------------------
bool Line::PointOnLine(double x, double y)
{
        return (fabs(A * x + B * y + C) < DOUBLE_EPS);
}

// --- public -----------------------------------------------------------------
double Line::dist(Cvector pt)
{
        return fabs((A*pt.x() + B*pt.y() + C) / sqrt(sqr(A) + sqr(B)));
}

// --- public -----------------------------------------------------------------
double Line::sqrdist(Cvector pt)
{
        return fabs(sqr(A*pt.x() + B*pt.y() + C) / (sqr(A) + sqr(B)));
}

// --- public -----------------------------------------------------------------
AngleDeg Line::angle()
{
        return ATan(-A / B);
}

// --- public -----------------------------------------------------------------
Cvector Line::ProjectPointUsingCircle(Cvector pt)
{
        /* here's the idea:
           first get the plane equation for the ray
           then use the closest distance formula to get the closest distance
           then using the eq for that line and the eq for a circle of the
           dist radius around our curretn position, find the nearest point.
           Whew! */

        /* compute the dist */
        Cvector retPt;
        double d = dist(pt); /* the min distance */

        /* intersect the circle and the line */
        double a,b,c; /* coefficent in quadratic to solve for x */
        double disc; /* discriminant in quadratic equation */
        a = 1 + sqr(A);
        b = 2 * (A * (pt.y() + C) - pt.x());
        c = sqr(pt.x()) + sqr(pt.y() + C) - sqr(d);
        disc = sqr(b) - 4 * a * c;
        /* the discriminant should be zero since this is the radius
           is the closest distance between the center and the line */
        if (fabs(disc) > DOUBLE_EPS)
                ROBOBUG1(ROBOBUG_ERROR,MATHS_LINE,
                         "GetClosestPointToBallPath: discrimannt is bad (%f)",
                         disc);
        retPt.setx(-b / (2 * a));

        /* we compute two possible solutions for y and then see which one
           is on the line of the ray's path */
        double sol1, sol2;
        sol1 = sqrt(sqr(d) - sqr(retPt.x() - pt.x()));
        sol2 = -sol1;
        sol1 += pt.y();
        sol2 += pt.y();

        if (fabs(A * (retPt.x()) + B * sol1 + C) < DOUBLE_EPS ) {
                /* sol1 is on line */
                retPt.sety(sol1);
        } else if (fabs(A * (retPt.x()) + B * sol2 + C) < DOUBLE_EPS ){
                /* sol2 is on line */
                retPt.sety(sol2);
        } else
                ROBOBUG(ROBOBUG_ERROR,MATHS_LINE,
                        "GetClosestPointToBallPath: neither solution works");

        ROBOBUG2(ROBOBUG_INFO,MATHS_LINE,
                 "dist: %f\t ptMod: %f\n",
                 d, sqrt(sqr(pt.x() - retPt.x()) + sqr(pt.y() - retPt.y())));

        return retPt;
}

// --- public -----------------------------------------------------------------
double Line::get_y(double x)
{
        if (B != 0) return (-A*x - C)/B;
        ROBOBUG(ROBOBUG_INFO,MATHS_LINE, "can't get y");
        return 0;
}

// --- public -----------------------------------------------------------------
double Line::get_x(double y)
{
        if (A != 0) return (-B*y - C)/A;
        ROBOBUG(ROBOBUG_INFO,MATHS_LINE, "can't get x");  
        return 0;
}

// --- public -----------------------------------------------------------------
bool Line::InBetween(Cvector pt, Cvector end1, Cvector end2)
{
        if (!OnLine(end1) || !OnLine(end2)) {
                ROBOBUG(ROBOBUG_INFO,MATHS_LINE,
                        "InBetween: passed in points that weren't on line");
        }
        
        pt = ProjectPoint(pt);
        double dist2 = end1.sqrdist(end2);
  
        return (pt.sqrdist(end1) <= dist2 && pt.sqrdist(end2) <= dist2) ?
                true : false;
}

// --- public -----------------------------------------------------------------
Cvector Line::intersection(Line l)
{
        Cvector result = 0;
        if (SameSlope(l)) {
                //if ( B == 0 && l.B == 0 || A/B == l.A/l.B ){
                ROBOBUG(ROBOBUG_ERROR,MATHS_LINE,
                        "Lines have same slope");
                return result;
        }

        if (B == 0) {
                result.set(-C/A, l.get_y(result.x()));
                return result;
        }

        if (l.B == 0) {
                result.set(-l.C/l.A, get_y(result.y()));
                return result;
        }

        result.set((C*l.B - B*l.C)/(l.A*B - A*l.B), get_y(result.x()));
        return result;
}  

// --- public -----------------------------------------------------------------
bool Line::IsPtCloserToPtOnLine(Cvector pt1, Cvector pt2, Cvector targ_pt)
{
        if (!OnLine(targ_pt))
                ROBOBUG(ROBOBUG_ERROR,MATHS_LINE,                
                        "IsPtCloserToPtOnLine: targ_pt not on line");

        pt1 = ProjectPoint(pt1);
        pt2 = ProjectPoint(pt2);

        return (pt1.dist(targ_pt) < pt2.dist(targ_pt)) ? true : false;  
}


// --- public -----------------------------------------------------------------
// return true on top/left part of plane
bool Line::HalfPlaneTest(Cvector pt)
{
        if (B==0) return (pt.x() < -C/A) ? true : false;
        return (pt.y() > get_y(pt.x())) ? true : false;
}

bool Line::SameSlope(Line l)
{
        return ( B == 0 && l.B == 0 || A/B == l.A/l.B ) ? true : false;    
}

// ----------------------------------------------------------------------------
// Rectangle ------------------------------------------------------------------
// most of this is CMU code

// --- public constructor -----------------------------------------------------
Rectangle::Rectangle()
{
        left_x = right_x = top_y = bottom_y = 0;
}

// --- public constructor -----------------------------------------------------
Rectangle::Rectangle(double l, double t, double r, double b) 
{
        RASSERT(r>l,ROBOBUG_ERROR,MATHS_RECTANGLE, "width must be positive");
        RASSERT(b>t,ROBOBUG_ERROR,MATHS_RECTANGLE, "height must be positive");

        left_x = l;     right_x = r;
        top_y = t;      bottom_y = b;
}

// --- public constructor -----------------------------------------------------
Rectangle::Rectangle(const Cvector &center, const Cvector &size)
{
        left_x = center.x() - size.x()/2.0;
        right_x = center.x() + size.x()/2.0;
        top_y = center.y() - size.y()/2.0;
        bottom_y = center.y() + size.y()/2.0;
}

// --- public -----------------------------------------------------------------
bool Rectangle::IsWithin(const Cvector &p)
{
        return ((p.x() >= left_x) && (p.x() <= right_x) &&
                (p.y() >= top_y) && (p.y() <= bottom_y));
}

// --- public -----------------------------------------------------------------
Cvector Rectangle::AdjustToWithin(const Cvector &p)
{
        Cvector r = p;
  
        if (r.y() - top_y < 0) r.sety(top_y);
        if (bottom_y - r.y() < 0) r.sety(bottom_y);
        if (right_x - r.x() < 0) r.setx(right_x);
        if (r.x() - left_x < 0) r.setx(left_x);

        return r;
}

// --- public -----------------------------------------------------------------
// order: top, right, bot, left
Line Rectangle::GetEdge(int n)
{
        switch (n % 4) {
        case 0: return TopEdge();
        case 1: return RightEdge();
        case 2: return BottomEdge();
        case 3: return LeftEdge();
        }

        ROBOBUG(ROBOBUG_ERROR,MATHS_RECTANGLE, "GetEdge:how did I get here?");
        return Line();
}

// --- public -----------------------------------------------------------------
//order: TL, TR, BR, BL
Cvector Rectangle::GetPoint(int n)
{
        switch (n % 4) {
        case 0: return TopLeftCorner();
        case 1: return TopRightCorner();
        case 2: return BottomRightCorner();
        case 3: return BottomLeftCorner();
        }    

        ROBOBUG(ROBOBUG_ERROR,MATHS_RECTANGLE, "GetPoint:how did I get here?");
        return Cvector(0,0);
}

// --- public -----------------------------------------------------------------
// find nearest horizontal line 
Cvector Rectangle::nearestHEdge(const Cvector &p)
{
        /* dangerous, if used by interrupt and main program (???) */
        static Cvector r;

        r.set(min(max(p.x(),left_x),right_x),
              ((p.y() - top_y) < (bottom_y - p.y())) ? top_y : bottom_y);

        return r ;
}

// --- public -----------------------------------------------------------------
// find nearest vertical line 
Cvector Rectangle::nearestVEdge(const Cvector &p)
{
        /* dangerous, if used by interrupt and main program (???) */
        static Cvector r;

        r.set(((p.x() - left_x) < (right_x - p.x())) ? left_x : right_x,
              min(max(p.y(),top_y), bottom_y));
        
        return r;
}

// --- public -----------------------------------------------------------------
Cvector Rectangle::nearestEdge(const Cvector &p) 
{
        if(min((p.x() - left_x), (right_x - p.x())) <
           min((p.y() - top_y), (bottom_y - p.y())))
                return nearestVEdge(p);
        else
                return nearestHEdge(p);
}

// --- public -----------------------------------------------------------------
Line Rectangle::nearestHEdgeLine(const Cvector& p) 
{
        return ((p.y() - top_y) < (bottom_y - p.y())) ?
                TopEdge() : BottomEdge() ;
}

// --- public -----------------------------------------------------------------
Line Rectangle::nearestVEdgeLine(const Cvector& p) 
{
        return ((p.x() - left_x) < (right_x - p.x())) ?
                LeftEdge() : RightEdge();
}

// --- public -----------------------------------------------------------------
Line Rectangle::nearestEdgeLine(const Cvector& p) 
{
        if(min((p.x()-left_x),(right_x-p.x())) <
           min((p.y()-top_y),(bottom_y-p.y())))
                return nearestVEdgeLine(p) ;
        else
                return nearestHEdgeLine(p) ;  
}

// --- public -----------------------------------------------------------------
double Rectangle::DistanceToEdge(const Cvector &p) 
{
        if (! IsWithin(p)) {
                Cvector q = AdjustToWithin(p);
                /* distance outside is a negative number */                
                return -(q.dist(p)); 
        }
        
        return min((p.x()-left_x),
                   min((right_x-p.x()),
                       min((p.y()-top_y),
                           (bottom_y-p.y()))));
}

// --- public -----------------------------------------------------------------
Cvector Rectangle::random() 
{
        /* dangerous, if used by interrupt and main program (???) */        
        static Cvector r;

        r.set(range_random(left_x, right_x), range_random(bottom_y, top_y));

        return r ;
}

// --- public -----------------------------------------------------------------
Rectangle Rectangle::expand(double val)
{
        return Rectangle(left_x - val, right_x + val,
                         top_y - val, bottom_y + val);
}

// --- public -----------------------------------------------------------------
void Rectangle::Print()
{
        printf("RECTANGLE:  x = %.1f to %.1f   y = %.1f to %.1f\n",
               LeftX(),RightX(),TopY(),BottomY());
}
