/* -*- mode: c++ -*-
   this file is part of the robolog libraries
   maths stuff definitions
   Copyright (C) Mar 22 1998 Oliver Obst
   $Id: maths.h,v 1.4 1999/05/05 14:09:08 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* parts of the code in this file are taken from
 * CMUnited98 (soccer client for Robocup98)
 * Peter Stone <pstone@cs.cmu.edu>
 * Computer Science Department
 * Carnegie Mellon University
 * Copyright (C) 1998 Peter Stone
 *
 * CMUnited-98 was created by Peter Stone, Manuela Veloso, and Patrick Riley
 *
 * You may copy and distribute this program freely as long as you retain
 * this notice.
 * If you make any changes or have any comments we would appreciate a message.
 */
#ifndef MATHS_H
#define MATHS_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>

class Cvector;
class Pvector;
class AngleRad;
class AngleDeg;
class Line;

const double m_deg2rad = M_PI / 180.0;
const double m_rad2deg = 180.0 / M_PI;

double deg2rad(double deg);
double rad2deg(double rad);

inline double max(const double &d1, const double &d2)
{ return d1 > d2 ? d1 : d2; }  

inline double min(const double &d1, const double &d2)
{ return d1 < d2 ? d1 : d2; }

inline double sqr(double d) 
{ return d * d; }

inline int sqr(int i)
{ return i * i; }

inline double abs(double d)
{ return ((d < 0) ? -d : d); }

inline int sgn(double d) 
{ return ((d < 0) ? -1 : 1); }


Cvector iexp(double d, bool deg=true);

int int_random(int n);
double range_random(double lo, double hi);

double normalize_angle(double x, bool deg=true);
double normalize_value(double x);

class AngleRad
{
public:
        AngleRad(double d=0);
        AngleRad(const AngleRad &ang);
        AngleRad(const AngleDeg &ang);
        
        AngleRad& normalize() { _normalize(); return *this; }

        double rad() const { return a; }
        double deg() const { return a * m_rad2deg; }

        AngleRad operator -();
        AngleRad operator +(const AngleRad &ang);
        AngleRad operator -(const AngleRad &ang);

        AngleRad operator =(const AngleRad &ang);
        AngleRad operator =(const AngleDeg &ang);
        AngleRad operator =(const double &d);
        AngleRad operator +=(const AngleRad &ang);
        AngleRad operator -=(const AngleRad &ang);
        AngleRad operator *=(const double &d);
        AngleRad operator /=(const double &d);

        int operator ==(const AngleRad &ang);
        int operator !=(const AngleRad &ang);

        operator double() { return a; }

protected:
        inline void _normalize();
        
        double a;
};


class AngleDeg
{
public:
        AngleDeg(double d=0);
        AngleDeg(const AngleDeg &ang);
        AngleDeg(const AngleRad &ang);
        
        AngleDeg& normalize() { _normalize(); return *this; }

        double rad() const { return a * m_deg2rad; }
        double deg() const { return a; }

        AngleDeg operator -();
        AngleDeg operator +(const AngleDeg &ang);
        AngleDeg operator -(const AngleDeg &ang);

        AngleDeg operator =(const AngleDeg &ang);
        AngleDeg operator =(const AngleRad &ang);
        AngleDeg operator =(const double &d);
        AngleDeg operator +=(const AngleDeg &ang);
        AngleDeg operator -=(const AngleDeg &ang);
        AngleDeg operator *=(const double &d);
        AngleDeg operator /=(const double &d);

        int operator ==(const AngleDeg &ang);
        int operator !=(const AngleDeg &ang);

        operator double() { return a; }

protected:
        inline void _normalize();
        
        double a;
};


inline double Sin(AngleDeg x) 
{ return sin(x.rad()); }

inline double Cos(AngleDeg x)
{ return cos(x.rad()); }

inline double Tan(AngleDeg x)
{ return tan(x.rad()); }

inline AngleDeg ASin(double x)
{ return (x >= 1) ? 90 : (x <= -1) ? -90 : rad2deg(asin(x)); }

inline AngleDeg ACos(double x)
{ return (x >= 1) ? 0 : (x <= -1) ? 180 : rad2deg(acos(x)); }

inline AngleDeg ATan(double x)
{ return rad2deg(atan(x)); }

inline AngleDeg ATan2(double x, double y)
{  return (x == 0 && y == 0) ? 0 : rad2deg(atan2(x,y)); }

inline double ACos(double x, bool deg=true)
{
        if (deg)
                return (x >= 1 ? 0 : (x <= -1 ? 180 : rad2deg(acos(x))));
        else
                return (x >= 1 ? 0 : (x <= -1 ? M_PI : acos(x)));
}

/**
 * A class for cartesian vectors. This class is an own implementation,
 * though it is very similar to the CMU vector class (Peter Stone)
 * and the turnballclient vector class (Alessandro Bissacco).
 * The global2relative and relative2global functions are copied
 * from the CMU vector class. 
 * The vector multiplication is different from CMU and turnballclient:
 * here it is like multiplying complex numbers and not the inner product.
 */
class Cvector
{
public:
        Cvector(double vx=0, double vy=0);
        Cvector(const Cvector &v);
        Cvector(double dist, AngleDeg dir);
        Cvector(double dist, AngleRad dir);
        
        Cvector& set(double vx=0, double vy=0);
        Cvector& set(double dist, AngleDeg dir);
        Cvector& set(double dist, AngleRad dir);
        inline void setx(double vx);
        inline void sety(double vy);
        
        inline double x() const;
        inline double y() const;
        double len() const;
        double sqrlen() const;
        double dir(bool deg=true) const;
        double dist(const Cvector &v1) const;
        double sqrdist(const Cvector &v1) const;

        void rotate(double angle, bool deg=true);
        Cvector rotate(AngleDeg ang) const;
        Cvector rotate(AngleRad ang) const;

        Cvector global2relative(Cvector orig, AngleDeg ang);
        Cvector global2relative(Cvector orig, AngleRad ang);
        Cvector relative2global(Cvector orig, AngleDeg ang);
        Cvector relative2global(Cvector orig, AngleRad ang);

        Cvector operator -() const;
        Cvector operator +(const Cvector &v) const;
        Cvector operator -(const Cvector &v) const;
        Cvector operator *(const Cvector &v) const;
        Cvector operator *(const double &d) const;
        Cvector operator /(const Cvector &v) const;
        Cvector operator /(const double &d) const;

        Cvector& operator =(const Cvector &v);
        Cvector& operator =(const double &d);
        Cvector& operator +=(const Cvector &v);
        Cvector& operator +=(const double &d);
        Cvector& operator -=(const Cvector &v);
        Cvector& operator -=(const double &d);
        Cvector& operator *=(const double &d);
        Cvector& operator /=(const double &d);

        friend Cvector operator *(const double &d, const Cvector &v);

        friend int operator ==(const Cvector &v, const Cvector &w);
        friend int operator ==(const Cvector &v, const double &d);
        friend int operator !=(const Cvector &v, const Cvector &w);
        friend int operator !=(const Cvector &v, const double &d);

        void print();
        
//        friend ostream& operator <<(ostream &os, const Cvector &v);

//        friend Cvector& operator =(const Polar &p);
//        friend Polar& operator =(const Cvector &v);

protected:
        double myx;
        double myy;
};


class Pvector
{
public:
        Pvector(double dst=0, double dir=0, bool deg=true);
        virtual ~Pvector();

        Pvector& set(double dst, double dir, bool deg=true);
        Pvector& set_cart(double vx, double vy);

        void    set_dist(const double dist);
        void    set_dir(const double dir, bool deg=true);
        
        double dist() const;
        double dir(bool deg=true);

        Pvector& operator =(const Pvector &v);

        friend int operator ==(Pvector &v0, Pvector &v1);


protected:
        double  mydist;
        double  dir_deg;
        double  dir_rad;
        bool    deg_valid;
        bool    rad_valid;
};


class Xvector : public Cvector
{
public:
        Xvector(double vx=0, double vy=0, bool auth=false);
        virtual ~Xvector();

        void set_authentic(bool auth=true);
        bool authentic() const;

protected:
        bool valid;
};


/**
 * a class representing lines. this is mostly taken from the
 * CMU line class (Peter Stone)
 */
class Line {
public:
        Line(double x_coef=0.0, double y_coef=0.0, double constant=0.0);
        void LineFromTwoPoints(Cvector pt1, Cvector pt2);
        void LineFromRay(Cvector orig, Cvector dir);
  
        bool PointOnLine(double x, double y);
        inline bool PointOnLine(Cvector pt)
                { return PointOnLine(pt.x(), pt.y()); }

        double dist(Cvector pt);
        double sqrdist(Cvector pt);
        AngleDeg angle();  // returns the angle of the line in [-90, 90]

        /**
         * Checks if a point is in the plane between two points of
         * the line.
         * @param pt            the point to check
         * @param end1          the first point on the line
         * @param end2          the second point on the line
         * @return              true if the perpendicular line
         *                      through pt intersects between end1 and end2
         */
        bool InBetween(Cvector pt, Cvector end1, Cvector end2);

        /* the buffer should really be linked to an option or something,
           but 1.0 is okay */
        inline bool OnLine(Cvector pt, double buffer = 1.0)
                { return (sqrdist(pt) < sqr(buffer)) ? true : false; }
  
        Cvector ProjectPointUsingCircle(Cvector pt);
        inline Cvector ProjectPoint(Cvector pt)
                { return intersection(perpendicular(pt)); }

        double get_y(double x);
        double get_x(double y);
        Cvector intersection(Line l);
        inline Line perpendicular(Cvector pt)
                { return Line(B,-A,A*pt.y() - B*pt.x()); }

        Line shift_y(double val)
                { return Line(A, B, C - val * B); }
        Line shift_x(double val)
                { return Line(A, B, C - val * A); }  

        /* returns whether the projection of pt1 is closer to targ_pt than the
           projection of pt2 */
        bool IsPtCloserToPtOnLine(Cvector pt1, Cvector pt2, Cvector targ_pt);
  
        bool HalfPlaneTest(Cvector pt); //return true on top/left part of plane
        bool SameSlope(Line l);
#if 0  
        friend ostream& operator <<(ostream & os, Line l)
                { return os << "#L("<<l.A<<", "<<l.B<<", "<<l.C<<")"; } 
        void Print(void) { cout << *this << endl; }
#endif        

private:
        double A,B,C; /* the three coeffs in the line equation */
        /* Ax + By + C = 0 */
};


inline Line LineFromTwoPoints(Cvector pt1, Cvector pt2)
{
        Line l;
        l.LineFromTwoPoints(pt1,pt2);
        return l;
}

inline Line LineFromRay(Cvector orig, Cvector dir)
{
        Line l;
        l.LineFromRay(orig,dir);
        return l;
}

inline Cvector GetClosestPointToRay(Cvector pt, Cvector orig, Cvector dir)
{
        Line l;
        l.LineFromRay(orig, dir);
        return l.ProjectPoint(pt);
}

/**
 * a class representing rectanglular areas. this is mostly taken from the
 * CMU rectangle class (Peter Stone)
 */
class Rectangle {
public:
        Rectangle();
        /*** !!! warning - alert - danger: changed the order of arguments! */
        Rectangle(double left, double top, double right, double bottom);
        Rectangle(const Cvector &center, const Cvector &size);

        int operator ==(const Rectangle &a) 
                { return (left_x == a.left_x) && (right_x == a.right_x) && 
                          (top_y == a.top_y) && (bottom_y == a.bottom_y); } 

        inline double TopY()     { return top_y; }
        inline double BottomY()  { return bottom_y; }
        inline double RightX()   { return right_x; }
        inline double LeftX()    { return left_x; }

        inline Cvector TopLeftCorner()
                { return Cvector(left_x, top_y); }
        inline Cvector TopRightCorner()
                { return Cvector(right_x, top_y); }
        inline Cvector BottomLeftCorner()
                { return Cvector(left_x, bottom_y); }
        inline Cvector BottomRightCorner()
                { return Cvector(right_x, bottom_y); }

        inline Line LeftEdge()
                { return LineFromTwoPoints(TopLeftCorner(),
                                           BottomLeftCorner()); }
        inline Line RightEdge()
                { return LineFromTwoPoints(TopRightCorner(),
                                           BottomRightCorner()); }
        inline Line TopEdge()
                { return LineFromTwoPoints(TopLeftCorner(),
                                           TopRightCorner()); }
        inline Line BottomEdge()
                { return LineFromTwoPoints(BottomLeftCorner(),
                                           BottomRightCorner()); }

        Line GetEdge(int n);  //order: top, right, bot, left; starting at 0
        Cvector GetPoint(int n); //order: TL, TR, BR, BL; starting at 0
  
        inline double Width()    { return right_x - left_x; }   
        inline double Height()   { return bottom_y - top_y; }  

        inline Cvector Center()
                { return Cvector( LeftX() + Width()/2, TopY() + Height()/2 ); }

        /**
         * Check if the point is in the rectangle.
         * @param p             the point to check
         * @return              true, if point is inside
         */
        bool   IsWithin(const Cvector& p);
        Cvector nearestHEdge(const Cvector& p);
        Cvector nearestVEdge(const Cvector& p);
        Cvector nearestEdge(const Cvector& p);

        /**
         * Returns the distance of a point to the nearest edge.
         * @param p             the point to check
         * @return              the distance (negative, if point outside)
         */
        double  DistanceToEdge(const Cvector& p);
        /**
         * Returns the nearest point within the rectangular area
         * @param p             the point to check
         * @return              if p is inside the rectangle,
         *                      p will be returned; else a point
         *                      on an edge of the rectangle is returned.
         */
        Cvector AdjustToWithin(const Cvector& p);

        Line nearestHEdgeLine(const Cvector& p);
        Line nearestVEdgeLine(const Cvector& p);
        Line nearestEdgeLine(const Cvector& p);
  
        Cvector random() ;

        Rectangle shrink(double val)
                { return expand(-val); }
        Rectangle expand(double val);

        Rectangle expandLeft(double val)
                {return Rectangle(left_x - val, right_x, top_y, bottom_y);}
        Rectangle expandRight(double val)
                {return Rectangle(left_x, right_x + val, top_y, bottom_y);}
        Rectangle expandTop(double val)
                {return Rectangle(left_x, right_x, top_y - val, bottom_y);}
        Rectangle expandBottom(double val)
                {return Rectangle(left_x, right_x, top_y, bottom_y + val);}
        Rectangle shrinkLeft(double val)
                { return expandLeft(-val); }
        Rectangle shrinkRight(double val)
                { return expandRight(-val); }
        Rectangle shrinkTop(double val)
                { return expandTop(-val); }
        Rectangle shrinkBottom(double val)
                { return expandBottom(-val); }

        void Print();

private:
        double left_x;
        double right_x;
        double top_y;
        double bottom_y;
};
#endif
