/* -*- mode: c++ -*-
   this file is part of the robolog libraries  
   Olivers Client Class Implementation
   Copyright (C) Oct 21 1997 Oliver Obst
   $Id: oclient.C,v 1.3 1999/05/05 14:27:55 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif


#include <signal.h>
#include <values.h>
#include <stdio.h>

#include "oclient.h"
#include "debug.h"

#ifndef O_ASYNC
# ifdef FASYNC
#  define O_ASYNC FASYNC
# else
#  define O_ASYNC 0
#  include <sys/filio.h>
#  include <sys/ioctl.h>
#  include <sys/sockio.h>
#  include <stropts.h>
# endif
#endif

#ifndef O_NONBLOCK
# define  O_NONBLOCK 0
#endif

#ifndef O_NDELAY
# define O_NDELAY 0
#endif

#if ! (O_NONBLOCK | O_NDELAY)
# warning Both O_NONBLOCK and O_NDELAY are unknown
#endif

// --- constructor ------------------------------------------------------------
// const char *hostname         name of the server
//       int  port              which port number to use
OClient::OClient(const char *hostname, int port)
{
        server = new char[ strlen(hostname) + 1 ];
        strcpy(server, hostname);
        portnr1 = port;
        socket2 = NO_SOCKET;
        log = false;
        alog = false;
}


// --- destructor -------------------------------------------------------------
OClient::~OClient()
{
        stop_server();
        stop_logfile();
        stop_audiologfile();
        
        delete server;
        OClient::close();
}


// --- public -----------------------------------------------------------------
// open connection to the server
// const char *hostname         name of the server
//       int  port              which port number to use
// return int                   socket number, if success; -1, if failure
int OClient::open(const char *hostname, int port)
{
        int socktype;               // Socket Typ

        // Server und Port festlegen
        // (bei Nullwerten sollen sie schon festgelegt sein)
        if (hostname != 0) {
                if (server != 0) {
                        delete server;
                        OClient::close();
                }
                server = new char[ strlen(hostname) + 1 ];
                strcpy(server, hostname);
        }
        if (port > 0)
                portnr1 = port;

        // die Adresse des Servers setzen
        memset( &serverAddr, 0, sizeof(serverAddr));
        serverAddr.sin_family = AF_INET;
        serverAddr.sin_port = htons(portnr1);
        
        // Server suchen und Adresse abspeichern
        if ((serverInfo = gethostbyname(server)) == 0) {
                // wurde Adresse numerisch angegeben ?
#ifdef HAVE_INET_ATON
                if (inet_aton(server,&serverAddr.sin_addr) == 0) {
                        ROBOBUG1(ROBOBUG_WARN,OCLIENT_SOCKET,
                                 "unknown host %s",server);
                        return -1;
                }
#else
                if ((serverAddr.sin_addr.s_addr = inet_addr(server))
                    == (unsigned long int) -1 ) {
                        ROBOBUG1(ROBOBUG_WARN,OCLIENT_SOCKET,
                                 "unknown host %s",server);
                        return -1;
                }
#endif HAVE_INET_ATON
        } else
                memcpy(&serverAddr.sin_addr,
                       serverInfo->h_addr,
                       serverInfo->h_length);

        sigset_t mask;
        sigemptyset(&mask);
        sigaddset(&mask, SIGIO);
        sigprocmask(SIG_BLOCK, &mask, 0);

        // UDP Socket erzeugen
        if ((sockfd1 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
                ROBOBUG1(ROBOBUG_WARN,OCLIENT_SOCKET,
                         "cannot create socket (%s)",strerror(errno));
                return -1;
        }

#if O_ASYNC        
        // this process should receive SIGIO signals
        if (fcntl(sockfd1, F_SETOWN, getpid()) != -1) {
                // set socket to non-blocking and asynchronous
                socktype = fcntl(sockfd1, F_GETFL, 0);
                // set at least one of NONBLOCK and NDELAY
                socktype |= O_NONBLOCK | O_NDELAY;
                socktype |= O_ASYNC;
                fcntl(sockfd1, F_SETFL, socktype);
        }
#else
      // this process should receive SIGIO signals        
      if (fcntl(sockfd1, F_SETOWN, getpid()) == 0) {
              // set socket to non-blocking
              socktype = fcntl(sockfd1, F_GETFL, 0);
                // set at least one of NONBLOCK and NDELAY
              socktype |= O_NONBLOCK | O_NDELAY;
              fcntl(sockfd1, F_SETFL, socktype);
              // set the socket state for asynchronous
              // ? do I have to set S_OUTPUT as well ??? at least, it works.
              ioctl(sockfd1, I_SETSIG, S_INPUT);
      }
#endif      
        return sockfd1;
}


// --- public -----------------------------------------------------------------
// close connection
void OClient::close()
{
        ::close(sockfd1);
}

// --- public -----------------------------------------------------------------
// read datagram from server
//      char *data              buffer to store received data
//      int  len                max. number of bytes to read
//                              (including terminating zero-byte)
// return int                   number of bytes read ( 0 .. len ) if success
//                              -1 if error
int OClient::receive(char *data, int len)
{
        int l;
        char tmpbuf[80];
        
        unsigned int slen = sizeof(serverAddr);
        unsigned int clen = sizeof(clientAddr);

        if (socket2 == SOCKET_CREATED) {
                l = recvfrom(sockfd2, tmpbuf, 80, 0,
                            (sockaddr *) &clientAddr, &clen);
                if (l > 0) {
                        socket2 = CONNECTION_ESTABLISHED;
                }
        }

        l = recvfrom(sockfd1, data, --len, 0, (sockaddr *) &serverAddr, &slen);
        
        if (l > 0) {
                data[ l ] = '\0';
                // send data to display client
                if (socket2 == CONNECTION_ESTABLISHED) {
                        sendto(sockfd2, data, l+1, 0,
                               (sockaddr *) &clientAddr, clen);
                }
                // write data to log file
                if (log) {
                        fprintf(logfile,"%s\n",data);
                        loglines++;
                        if (loglines == maxloglin) {
                                fflush(logfile);
                                loglines = 0;
                        }
                }
                // write audio data to audio log file
                if (alog && data[1] == 'h') {
                        fprintf(alogfile,"%s\n",data);
                        aloglines++;
                        if (aloglines == maxaloglin) {
                                fflush(alogfile);
                                aloglines = 0;
                        }
                }
        } else {
                if (l == -1 && errno == EWOULDBLOCK) {
                        l = 0;
                        data[ 0 ] = '\0';
                } else {
                        l = -1;
                }
        }
        
        return l;
}

// --- public -----------------------------------------------------------------
// send a string to server
int OClient::send(const char *data)
{
        if (!data) return 0;
        
        int len = strlen(data);

        if (log) fprintf(logfile,">> %s\n\n",data);
        
        if (sendto(sockfd1, data, len, 0,
                   (sockaddr *) &serverAddr, sizeof(serverAddr)) != len)
                return -1;
        else
                return 0;
}


// --- public -----------------------------------------------------------------
// send ``len'' bytes to server
int OClient::send(const char *data, int len)
{
        if (!data) return 0;

        if (log) fprintf(logfile,">> %s\n\n",data);
        
        if (sendto(sockfd1, data, len, 0,
                   (sockaddr *) &serverAddr, sizeof(serverAddr)) != len)
                return -1;
        else
                return 0;
}

// --- public -----------------------------------------------------------------
// Start a UDP-Server which sends out the data the agent receives.
int OClient::start_server(int offset, int base=7000)
{
        if (socket2 != NO_SOCKET) return 1;
        
        portnr2 = base + offset;
        
        memset(&clientAddr, 0, sizeof(clientAddr));
        clientAddr.sin_family = AF_INET;
        clientAddr.sin_addr.s_addr = INADDR_ANY;
        clientAddr.sin_port = htons(portnr2);
        
        if ((sockfd2 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
                ROBOBUG1(ROBOBUG_WARN,OCLIENT_SERVER,
                         "cannot create socket (%s)",strerror(errno));
                return -1;
        }

        if (bind(sockfd2, (sockaddr *) &clientAddr, sizeof(clientAddr)) < 0) {
                ROBOBUG1(ROBOBUG_WARN,OCLIENT_SERVER,
                         "bind failed (%s)",strerror(errno));
                return -1;
        }

        socket2 = SOCKET_CREATED;
        
        return 0;
}

// --- public -----------------------------------------------------------------
// Stop the UDP-Server which sends out the data the agent receives.
void OClient::stop_server()
{
        if (socket2 != NO_SOCKET) {
                socket2 = NO_SOCKET;
                ::close(sockfd2);
        }
        return;
}

// --- public -----------------------------------------------------------------
// Start writing logfile.
int OClient::start_logfile(const char *filename, bool append, int flushfreq)
{
        if (!filename) return -1;
        
        maxloglin = (flushfreq >=0) ? flushfreq : 1;
        loglines = 0;

        if (append) logfile = fopen(filename,"a");
        else logfile = fopen(filename,"w");

        if (!logfile) {
                log = false;
                return -1;
        }
        
        fprintf(logfile,"# robolog-%s logfile\n",VERSION);        
        log = true;
        return 0;
}

// --- public -----------------------------------------------------------------
// Stop writing logfile.
void OClient::stop_logfile()
{
        if (log) fclose(logfile);
        log = false;
}

// --- public -----------------------------------------------------------------
// Start writing audio logfile.
int OClient::start_audiologfile(const char *filename, bool append,
                                int flushfreq) 
{
        if (!filename) return -1;
        
        maxaloglin = (flushfreq >= 0) ? flushfreq : 1;
        aloglines = 0;

        if (append) alogfile = fopen(filename,"a");
        else alogfile = fopen(filename,"w");

        if (!alogfile) {
                alog = false;
                return -1;
        }
                
        fprintf(alogfile,"# robolog-%s audiologfile\n",VERSION);        
        alog = true;
        return 0;
}

// --- public -----------------------------------------------------------------
// Stop writing audio logfile.
void OClient::stop_audiologfile()
{
        if (alog) fclose(alogfile);
        alog = false;
}
