/* -*- mode: c++ -*-
   
   this file is part of the robolog libraries
   Copyright (C) Oct 21 1997 Oliver Obst
   basic functions for client programs using UDP
   $Id: oclient.h,v 1.3 1999/05/05 14:25:50 frvit Exp $

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/ 
#ifndef OCLIENT_H
#define OCLIENT_H

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <errno.h>
#include <stdio.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

/**
 * An UDP/IP client class. This client connects to a given host on a given
 * port.
 *
 * Either give the hostname and portname while initializing the client, or
 * give hostname and portnumber if opening the connection.
 * The hostname can be a symbolic name (like "www.uni-koblenz.de") or an
 * IP-Address (e.g. "141.26.4.18").
 *
 * opening the connection as well as sending and receiving data return -1,
 * if the corresponding function does not succeed.
 *
 * @author      Oliver Obst <frvit@uni-koblenz.de>
 * @version     $Id: oclient.h,v 1.3 1999/05/05 14:25:50 frvit Exp $
 * @short       An UDP/IP client class.
 */
class OClient
{
public:
        /**
         * Creates a client and saves a host name and a port number.
         *
         * If you give a host name and a port number, you don't
         * have do give them again when calling <code>open()</code>.
         * The host name is the name of the server, the client will
         * connect to.
         *
         * @param hostname      name or IP of the server to connect to.
         * @param port          port number to use.
         */
        OClient(const char *hostname=0, int port=0);
        /**
         * Destructor. Closes the socket, if it is still open.
         */
        virtual ~OClient();

        /**
         * Creates a socket and opens a connection to the server.
         *
         * This is your last chance to give a server name (or IP) and a
         * port number.
         * If <code>hostname == 0</code>, the host name is assumed to be known.
         * If <code>port == 0</code>, the port number is assumed to be known.
         *
         * @param hostname      name or IP of the server to connect to
         * @param port          port number to use
         * @return              socket number if success; -1 if failure
         */
        int     open(const char *hostname=0, int port=0);
        /**
         * Simply closes the socket and returns nothing.
         */
        void    close();
        /**
         * Sends a null terminated string to the server. The null byte itself
         * will not be transmitted.
         * @param data          a C/C++ string
         * @return              0 if success, -1 if failure
         */
        int     send(const char *data);
        /**
         * Sends a block of data of a given size to the server.
         * @param data          the bytes you want to send
         * @param len           number of bytes to send
         * @return              0 if success, -1 if failure
         */
        int     send(const char *data, int len);
        /**
         * Reads a datagram from the server.
         * @param data          a buffer to store received data
         * @param len           size of the buffer
         * @return              number of bytes read (<code>0..len</code>),
         *                      if success; -1 if error
         */
        int     receive(char *data, int len);
        /**
         * Start a UDP-Server which sends out the data the agent receives.
         * This is based on an idea from ATH.
         * @param offset        will be added to the base port number
         * @param base          the base number for the port the server will
         *                      use.
         * @return              0 if success, -1 if error,
         *                      1 if socket already exists
         */
        int     start_server(int offset, int base=7000);
        /**
         * Close the UDP-Server.
         */
        void    stop_server();
        /**
         * Start writing logfile.
         * This idea is from the CMU'98 implementation.
         * @param filename      name of logfile
         * @param append        append to old file
         * @param flushfreq     flush frequency
         * @return              0 if success, -1 if error
         */
        int     start_logfile(const char *filename,
                              bool append = false,
                              int flushfreq = 50);
        /**
         * Stop writing logfile.
         */
        void    stop_logfile();
        /**
         * Start writing audio logfile.
         * This idea is from the CMU'98 implementation.
         * @param filename      name of logfile
         * @param append        append to old file
         * @param flushfreq     flush frequency
         * @return              0 if success, -1 if error
         */
        int     start_audiologfile(const char *filename,
                                   bool append = false,
                                   int flushfreq = 50);
        /**
         * Stop writing audio logfile.
         */
        void    stop_audiologfile();
        
protected:
        enum socket_status { NO_SOCKET,
                             SOCKET_CREATED,
                             CONNECTION_ESTABLISHED };
        
        char            *server;                // server string
        int             portnr1;                // soccer server port number
        sockaddr_in     serverAddr;             // soccer server address
        hostent         *serverInfo;            // info about soccer server
        int             sockfd1;                // socket for soccer server

        int             portnr2;                // display client port number
        sockaddr_in     clientAddr;             // display client address
        int             sockfd2;                // socket for display client

        socket_status   socket2;                // display client socket status

        FILE            *logfile;               // stream for the log
        FILE            *alogfile;              // stream for audio log
        bool            log;                    // flag if to write logfile
        bool            alog;                   // flag if to write audio log
        int             maxloglin;              // log lines / flush
        int             maxaloglin;             // audio log lines / flush
        int             loglines;               // log lines written
        int             aloglines;              // audio log lines written
};
#endif /* OCLIENT_H */
