/* -*- mode: c++ -*-
   
   this file is part of the robolog libraries
   Copyright (C) Oct 21 1997 Oliver Obst
   ring class to store objects for a time
   $Id: ring.C,v 1.2 1999/05/05 14:32:49 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "constants.h"   
#include "ring.h"

ObjRing::ObjRing(int rsize)
{
        size = rsize;
        objects = new ObjArray[size + 2];
        keys = new int[size];
        first = -1;
        last = -1;
        entries = 0;
        empty = size + 1;
}

ObjRing::~ObjRing()
{
        delete[] objects;
        delete[] keys;
}

int ObjRing::inc(int i) const
{
        return ((++i) % size);
}        

int ObjRing::dec(int i) const
{
        return ((--i) % size);
}

void ObjRing::reset()
{
        first = -1;
        last = -1;
        entries = 0;
}

void ObjRing::store(const ObjArray &obj, int key)
{
        if (first >= 0 && keys[first] > key) reset(); // should not happen
        first = inc(first);
        if (entries == 0 || entries == size) last = inc(last);
                
        if (entries < size) entries++;
        objects[first] = obj;
        keys[first] = key;
}

ObjArray& ObjRing::peek(int key) const
{
        int i = first;

        if (i >= 0) {
                if (key < 0 || key > keys[first]) key = keys[first];
                if (key < keys[last]) i = last;
                else while (i != last && key < keys[i]) i = dec(i);
        }
        else i = empty;
        
        return objects[i];
}

bool ObjRing::peek(ObjArray &obj, int &key) const
{
        bool found = true;
        int i = first;

        if (i < 0) {
                key = -1;
                return false;
        }
        
        if (key < 0) key = keys[first];
        
        if (key > keys[first]) {
                key = keys[first];
                found = false;
        }
        else if (key < keys[last]) {
                key = keys[last];
                found = false;
                i = last;
        }
        while (i != last && key < keys[i]) i = dec(i);
        found = (key == keys[i]);
        obj = objects[i];
        key = keys[i];
        return found;
}

// ObjArray& ObjRing::operator[](int i) const
// {
//         if (peek(objects[size], i)) return objects[size];
//         else return objects[size + 1];
// }

// ----------------------------------------------------------------------------
MsgRing::MsgRing(int rsize)
{
        size = rsize;
        objects = new MsgArray[size + 2];
        keys = new int[size];
        first = -1;
        last = -1;
        entries = 0;
        nextkey = -1;
        empty = size + 1;
}

MsgRing::~MsgRing()
{
        delete[] objects;
        delete[] keys;
}

int MsgRing::inc(int i) const
{
        return ((++i) % size);
}        

int MsgRing::dec(int i) const
{
        return ((--i) % size);
}

void MsgRing::reset()
{
        first = -1;
        last = -1;
        entries = 0;
        nextkey = -1;
}

void MsgRing::inc_nextkey()
{
        int next = inc(nextkey);
        
        if (last <= first) {
                if (next <= first) nextkey = next;
        } else {
                if (next > last || next <= first) nextkey = next;
        }
}

void MsgRing::store(MsgArray obj, int key)
{
        if (first >= 0 && key <= keys[first]) {
                int i;
                for (i=0; i < obj.get_entries(); i++) {
                        objects[first].add_message(obj[i]);
                }
        } else {
                first = inc(first);
                if (entries == 0 || entries == size) {
                        if (last == nextkey) nextkey = inc(nextkey);
                        last = inc(last);
                }
                if (entries < size) entries++;
                objects[first] = obj;
                keys[first] = key;
        }
}

MsgArray& MsgRing::peek(int key) const
{
        int i = first;

        if (i >= 0) {
                if (key < 0 || key > keys[first]) key = keys[first];
                if (key < keys[last]) i = last;
                else while (i != last && key < keys[i]) i = dec(i);
        }
        else i = empty;
        
        return objects[i];
}

bool MsgRing::peek(MsgArray &obj, int &key) const
{
        bool found = true;
        int i = first;

        if (i < 0) {
                key = -1;
                return false;
        }
        
        if (key < 0) key = keys[first];
        
        if (key > keys[first]) {
                key = keys[first];
                found = false;
        }
        else if (key < keys[last]) {
                key = keys[last];
                found = false;
                i = last;
        }
        while (i != last && key < keys[i]) i = dec(i);
        found = (key == keys[i]);       // flasch
        obj = objects[i];
        key = keys[i];
        return found;
}

// MsgArray& MsgRing::operator[](int i) const
// {
//         cerr << "~ operator [] " << i << "\n";
//         if (peek(objects[size], i)) return objects[size];
//         else return objects[empty];
//         cerr << "~ operator [] finished\n";
// }
#if 0
ostream& operator <<(ostream &os, const MsgRing &mr)
{
        int i;
        if (mr.last!=-1) {
                os << "[";
                i=mr.last;

                while (i != mr.first) {
                        os << (mr).objects[i] << ", ";
                        i = mr.inc(i);
                }
                os << mr.objects[mr.first] << "]";
        }
        return os;
}
#endif        

// ----------------------------------------------------------------------------
KnowRing::KnowRing(int rsize)
{
        size = rsize;
        objects = new Knowledge[size + 2];
        keys = new int[size];
        first = -1;
        last = -1;
        entries = 0;
        empty = size + 1;
}

KnowRing::~KnowRing()
{
        delete[] objects;
        delete[] keys;
}

int KnowRing::inc(int i) const
{
        return ((++i) % size);
}        

int KnowRing::dec(int i) const
{
        return ((--i) % size);
}

void KnowRing::reset()
{
        first = -1;
        last = -1;
        entries = 0;
}

void KnowRing::store(const Knowledge &obj, int key)
{
        if (first >= 0 && keys[first] > key) reset(); // should not happen
        first = inc(first);
        if (entries == 0 || entries == size) last = inc(last);
        if (entries < size) entries++;
        objects[first] = obj;
        keys[first] = key;
}

Knowledge& KnowRing::peek(int key) const
{
        int i = first;

        if (i >= 0) {
                if (key < 0 || key > keys[first]) key = keys[first];
                if (key < keys[last]) i = last;
                else while (i != last && key < keys[i]) i = dec(i);
        }
        else i = empty;
        
        return objects[i];
}

bool KnowRing::peek(Knowledge &obj, int &key) const
{
        bool found = true;
        int i = first;

        if (i < 0) {
                key = -1;
                return false;
        }
        
        if (key < 0) key = keys[first];
        
        if (key > keys[first]) {
                key = keys[first];
                found = false;
        }
        else if (key < keys[last]) {
                key = keys[last];
                found = false;
                i = last;
        }
        while (i != last && key < keys[i]) i = dec(i);
        found = (key == keys[i]);
        obj = objects[i];
        key = keys[i];
        return found;
}

// Knowledge& KnowRing::operator[](int i) const
// {
//         if (peek(objects[size], i)) return objects[size];
//         else return objects[size + 1];
// }
        

// // --- constructor ------------------------------------------------------------
// template<class T, int n>
// Ring<T,n>::Ring(int rsize)
// {
//         size = rsize;
//         objects = new (T(n))[size + 1];
//         keys = new int[size];
//         first = -1;
//         last = -1;
//         entries = 0;
// }

// template<class T, int n>
// Ring<T,n>::~Ring()
// {
//         delete[] objects;
//         delete[] keys;
// }

// template<class T, int n>
// int Ring<T,n>::inc(int i)
// {
//         return ((++i) % size);
// }        


// template<class T, int n>
// int Ring<T,n>::dec(int i)
// {
//         return ((--i) % size);
// }


// template<class T, int n>
// void Ring<T,n>::reset()
// {
//         first = -1;
//         last = -1;
//         entries = 0;
// }


// template<class T, int n>
// void Ring<T,n>::store(const T &obj, int key)
// {
//         if (first >= 0 && keys[first] > key) reset(); // should not happen
//         first = inc(first);
//         if (entries == 0 || entries == size) last = inc(last);
//         if (entries < size) entries++;
//         objects[first] = obj;
//         keys[first] = key;
// }


// template<class T, int n>
// bool Ring<T,n>::peek(T &obj, int &key)
// {
//         bool found = true;
//         int i = first;

//         if (i < 0) {
//                 key = -1;
//                 return false;
//         }
        
//         if (key < 0) key = keys[first];
        
//         if (key > keys[first]) {
//                 key = keys[first];
//                 found = false;
//         }
//         else if (key < keys[last]) {
//                 key = keys[last];
//                 found = false;
//                 i = last;
//         }
//         while (i != last && key < keys[i]) i = dec(i);
//         found = (key == keys[i]);
//         obj = objects[i];
//         key = keys[i];
//         return found;
// }

// template<class T, int n>
// T& Ring<T,n>::operator[](int i)
// {
//         T x;
//         if (peek(x, i)) return x;
//         else return objects[size];
// }
