/* -*- mode: c++ -*-
   
   this file is part of the robolog libraries
   Copyright (C) Oct 21 1997 Oliver Obst
   template ring class to store things for a time
   $Id: ring.h,v 1.2 1999/05/05 14:31:52 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _RING_H
#define _RING_H

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "objarray.h"
#include "msgarray.h"
#include "knowledge.h"

class ObjRing
{
public:
        ObjRing(int rsize);
        ~ObjRing();

        void    store(const ObjArray &obj, int key);
        ObjArray& peek(int key) const;
        bool    peek(ObjArray &obj, int &key) const;
        void    reset();
        inline  int get_size() const { return size; }
        inline  int get_entries() const { return entries; }
//        ObjArray& operator[](int i) const;

private:
        int     inc(int i) const;
        int     dec(int i) const;
        
        int     first;
        int     last;
        int     size;
        int     entries;
        int     empty;
        
        ObjArray *objects;
        int     *keys;
};

class MsgRing
{
public:
        MsgRing(int rsize);
        ~MsgRing();

        void    store(MsgArray obj, int key);
        MsgArray& peek(int key) const;
        bool    peek(MsgArray &obj, int &key) const;
        void    reset();
        void    inc_nextkey();
        inline  int get_size() const { return size; }
        inline  int get_entries() const { return entries; }
        inline  int get_nextkey() const
                { return (nextkey < 0 ? nextkey : keys[nextkey]); }
        
//        MsgArray& operator[](int i) const;
#if 0
        friend ostream& operator <<(ostream &os, const MsgRing &mr);
#endif
        
private:
        int     inc(int i) const;
        int     dec(int i) const;
        
        int     first;
        int     last;
        int     size;
        int     entries;
        int     nextkey;
        int     empty;
        
        MsgArray *objects;
        int     *keys;
};

class KnowRing
{
public:
        KnowRing(int rsize);
        ~KnowRing();

        void    store(const Knowledge &obj, int key);
        Knowledge& peek(int key) const;
        bool    peek(Knowledge &obj, int &key) const;
        void    reset();
        inline  int get_size() const { return size; }
        inline  int get_entries() const { return entries; }
        inline  int last_key() const { return (first < 0 ? 0 : keys[first]); }
        
//        Knowledge& operator[](int i) const;

private:
        int     inc(int i) const;
        int     dec(int i) const;
        
        int     first;
        int     last;
        int     size;
        int     entries;
        int     empty;
        
        Knowledge *objects;
        int     *keys;
};

        
// /**
//  * A Ring to store things for a while.
//  *
//  * The memory needed to store the objects will be created when calling
//  * the ring constructor. After storing <em>size of ring</em> objects, the
//  * oldest objects will be overwritten.
//  *
//  * @author      Oliver Obst <frvit@uni-koblenz.de>
//  * @version     $Id: ring.h,v 1.2 1999/05/05 14:31:52 frvit Exp $
//  * @short       A Ring to store objects for a while.
//  **/
// template<class T, int n>
// class Ring
// {
// public:
// /**
//  * Creates a ring with <code>rsize</code> slots to store objects.
//  *
//  * @param rsize         number of objects to keep max.
//  **/
//         Ring(int rsize);
// /**
//  * Destructor. The allocated memory is freed.
//  **/
//         ~Ring();

// /**
//  * store object <code>obj</code> with time stamp <code>key</code>. resets the
//  * ring if <code>key</code> is less then the latest object.
//  *
//  * @param obj   object to store
//  * @param key   time stamp
//  **/
//         void    store(const T &obj, int key);

// /**
//  * Request object with time stamp <code>key</code>. A negative key requests the
//  * latest object, which will be returned in <code>obj</code>. <code>key</code>
//  * is set to the time stamp of the latest object and <code>peek</code>
//  * returns <code>true</code>.
//  *
//  * If <code>key</code> is non-negative, <code>peek</code> returns true, if an
//  * object with time stamp <code>key</code> was exactly found. <code>obj</code>
//  * returns the requested object.
//  *
//  * If no object with time stamp <code>key</code> can be found,
//  * <code>peek</code> returns false and an approximate value for
//  * <code>key</code> and <code>obj</code>. The approximation works like this:
//  * If <code>key</code> is higher than the highest stamp, the latest (highest)
//  * object/key is returned. If <code>key</code> is smaller then the smallest
//  * stamp, the oldest (smallest) object/key is returned. If the key is in
//  * between the latest and the oldest key, but not found exactly, the next
//  * older object/key will be returned.
//  *
//  * @param obj   requested object
//  * @param key   time stamp
//  * @return      true if an object with time stamp key was found
//  **/        
//         bool    peek(T &obj, int &key);

// /**
//  * reset the ring (i.e. set the number of elements to 0)
//  **/
//         void    reset();

// /**
//  * return size of the ring
//  *
//  * @return      number of slots for objects
//  **/
//         inline  int get_size() { return size; }

// /**
//  * return number of used slots
//  *
//  * @return      number of entries
//  **/
//         inline  int get_entries() { return entries; }

// /**
//  * return object with stamp key
//  *
//  * @return      object
//  **/
//         T&      operator[](int i);

// private:
//         int     inc(int i);
//         int     dec(int i);
        
//         int     first;
//         int     last;
//         int     size;
//         int     entries;
        
//         T       *objects;
//         int     *keys;
// };
  
#endif /* _RING_H */
