/* -*- mode: c++ -*-
   
   this file is part of the robolog libraries
   Copyright (C) Dec 01 1997 Oliver Obst
   methods for robot self localisation
   $Id: self.h,v 1.5 1999/05/05 14:38:42 frvit Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _SELF_H
#define _SELF_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>

#include "sfield.h"
#include "objarray.h"
#include "knowledge.h"

class Knowledge;

/**
 * A class used for robot position estimation.
 *
 * The robot position is a tuple <em>(x, y, th)</em> of real numbers,
 * where <em>x</em> and <em>y</em> represent the coordinates on the soccer
 * field and <em>th</em> represents the robots orientation (heading).
 *
 * The methods in this class get their input data by sensor information
 * of the client, provided by the @ref Knowledge class.
 *
 * @author      Oliver Obst <frvit@uni-koblenz.de>
 * @version     $Id: self.h,v 1.5 1999/05/05 14:38:42 frvit Exp $
 * @short       A class used for robot position estimation.
 */
class Self : public Sfield
{
public:
        /**
         * Creates a position estimator class.
         *
         * @param s              a pointer to an initialized knowledge base
         */
        Self(ObjArray *oa0 = 0, Knowledge *k0 = 0);
        /**
         * Destructor.
         */
        virtual ~Self();

        /**
         * Calculates possible clients poses <em>(x, y, th)</em> by calling
         * @ref Self#intersections , @ref Self#linecheck , @ref Self#selection
         * and @ref Self#direction . After all, the (hopefully) right one of
         * the calculated poses is estimated by @ref Objarray#calc_abs .
         */
        void    locate();
        /**
         * Calculates direction (heading) of the client. This implementation
         * works only if the clients position is already known.
         * Could also work if regarding the lines on the field (but it does
         * not yet)
         */
        void    direction();
        
        void    intersections();
        void    linecheck();
        void    linecalc(int i);
        void    selection();

        /**
         * Change estimated direction (heading) of the client. This information
         * will be used if the client doesn't see enough landmarks.
         **/
        void    turn(double angle);

        /**
         * Change estimated position of the client. This information
         * will be used if the client doesn't see enough landmarks.
         **/
        void    moveto(double x, double y);

        /**
         * Decrease remembrance value of "odometer" information.
         **/
        void    forget();

protected:
        void hq_point_line(ObjInfo &point, ObjInfo &line,
                           Cvector &pos, double &theta);
        void hq_point_point(ObjInfo &point1, ObjInfo &point2,
                            Cvector &pos, double &theta);
        void lq_point_point(ObjInfo &p1, ObjInfo &p2,
                            Cvector &pos1, Cvector &pos2);

        void locate_2p(ObjInfo &p0, ObjInfo &p1, Cvector &p, double &theta);
        void locate_3p(ObjInfo &p0, ObjInfo &p1, ObjInfo &p2,
                       Cvector &p, double &theta);
        
        double  sqrdist(Cvector v0, Cvector v1);
        int     fuzzcmp(double d0, double d1);
        
        Cvector linepos[ 2 ][ MaxObject ];
        Xvector intersec[ 2 ][ MaxObject ];
        ObjArray *oa;
        Knowledge *k;
        
        int     insects;
        int     definite;
        int     lp;

        Cvector pos;
        Cvector lastpos;
        Cvector odopos;
        
        double  angle, headangle, bodyangle;
        double  lastangle;
        double  odoangle;

        int     remember;
};
#endif
